// ==============================================================
// Name:		x2c - MSXMLImport.cpp
//
// Description:	interfaces to MSXML com object
//
// Copyright (C) 2001-2008, J. Martin Wehlou, Man in The Middle AB
// martin@wehlou.com, martin@mitm.se, MiTM AB, Flintstensvagen 10, 75267 Uppsala, Sweden
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// $LastChangedDate: $
// $LastChangedRevision: $
// $LastChangedBy: $
// $HeadURL: $
// $Id: $
// ==============================================================

//#include "stdafx.h"
#include "MSXMLImport.h"

#include <sstream>
#include <string>

using std::wstring;

namespace m {

///////////////////////////////////////////////////////////////////////////
// Function: set attribute on given node, with and without the kitchen sink
///////////////////////////////////////////////////////////////////////////

void SetAttribute(XML_NodePtr pNode, CComBSTR bstrName, int iValue)
{
	CComVariant	vValue = (int)iValue;
	SetAttribute(pNode, bstrName, vValue);
}

void SetAttribute(XML_NodePtr pNode, CComBSTR bstrName, CComBSTR bstrValue)
{
	CComVariant vValue = (BSTR)bstrValue;
	SetAttribute(pNode, bstrName, vValue);
}

void SetAttribute(XML_NodePtr pNode, CComBSTR bstrName, CComVariant vValue)
{
	// for the callers' simplicity's sake, we allow null variants and empty, and simply ignore those

	if (pNode.p && vValue.vt != VT_NULL && vValue.vt != VT_EMPTY) {
		// get the 'element' interface on the node
		XML_ElementPtr pEle;
		pNode->QueryInterface(&pEle);
		pEle->setAttribute(bstrName.m_str, vValue);
	}
}

void SetAttribute(XML_NodePtr pNode, const wstring& sName, const wstring& sValue)
{
	SetAttribute(pNode, CComBSTR(sName.c_str()), CComBSTR(sValue.c_str()));
}


// ==========================================================================
//  function: remove attribute
// ==========================================================================

void DelAttribute(XML_NodePtr pNode, const wstring& sName)
{
	XML_ElementPtr spElement;
	pNode->QueryInterface(&spElement);
	ATLASSERT(spElement.p);
	spElement->removeAttribute(_bstr_t(sName.c_str()));
}


///////////////////////////////////////////////////////////////////////////
// Function: get the (only or first) node at any depth from given root
// that has the given attribute with the given value
///////////////////////////////////////////////////////////////////////////

XML_NodePtr GetNodeOnAttrib(XML_NodePtr pRoot, CComBSTR bstrName, void* ptr)
{
	// make a variant rep of the pointer
	CComVariant vValue = (long)ptr;

	return GetNodeOnAttrib(pRoot, bstrName, vValue);
}

XML_NodePtr GetNodeOnAttrib(XML_NodePtr pRoot, CComBSTR bstrName, CComVariant vValue)
{
	// since we need a string for the select string, convert the variant if it's something else
	if (vValue.vt != VT_BSTR)
		vValue.ChangeType(VT_BSTR);

	// construct the select string from attrib name and value
	// will look something like:
	//      "//*[@HTREEITEM=12345678]"
	// ... where //* means 'any node descendant at any depth, and [@...] means 'with attribute...'

	std::wstringstream wss;
	wss << L"//*[@" << bstrName.m_str << L"=" << vValue.bstrVal << L"]";
	CComBSTR bstrSelect (wss.str().c_str());

	// now go grab that node (result may be NULL if none found)
	XML_NodePtr node = pRoot->selectSingleNode(bstrSelect.m_str);

	return node;
}

////////////////////////////////////////////////////////////////////////////
// Function: get text value of (first or only) childnode with given name
////////////////////////////////////////////////////////////////////////////

CComVariant GetChildNodeText(XML_NodePtr pParent, CComBSTR bstrChildNodeName)
{
	XML_NodePtr node = pParent->selectSingleNode(bstrChildNodeName.m_str);
	
	CComVariant var;

	if (node)
	{
		var = (BSTR)node->Gettext();
	}
	else
	{
		var.vt = VT_EMPTY;
	}
	return var;
}

////////////////////////////////////////////////////////////////////////////////
// retrieve selected node (first or only), return attribute as bstr in variant
////////////////////////////////////////////////////////////////////////////////

// ------------------------------------------------------------------------- //

wstring GetSelectedNodeAttribute(XML_NodePtr pParent, const wstring& sSelect, const wstring& sAttribName)
{
	XML_NodePtr spNode = pParent->selectSingleNode(_bstr_t(sSelect.c_str()));
	if (spNode)
		return GetAttribute(spNode, sAttribName);
	else
		return L"";
}

// ------------------------------------------------------------------------- //

wstring GetNodeName(XML_NodePtr spNode)
{
	_bstr_t bstr = spNode->GetnodeName();
	return static_cast<wchar_t*>(bstr);	// _bstr_t does the magic
}

// ------------------------------------------------------------------------- //

wstring	GetAttribute(XML_NodePtr pNode, const wstring& sAttrName)
{
//	USES_CONVERSION;

//	wstring wsAttrName = A2W(sAttrName.c_str());
	XML_NamedNodeMapPtr nnMap;
	XML_NodePtr attr;
	
	if ((pNode) &&
		(nnMap = pNode->Getattributes()) &&
		(attr = nnMap->getNamedItem(sAttrName.c_str())))
	{
		return (wchar_t*)attr->Gettext();
	}
	else
	{
		return L"";
	}
}


////////////////////////////////////////////////////////////////////////////////
// Function: add child node, or update its contents
////////////////////////////////////////////////////////////////////////////////

void SetChildNodeText(XML_NodePtr pParent, CComBSTR bstrChildNodeName, CComVariant vValue)
{
	if (vValue.vt != VT_NULL)
	{
		if (vValue.vt != VT_BSTR)
			vValue.ChangeType(VT_BSTR);

		XML_NodePtr node = pParent->selectSingleNode(bstrChildNodeName.m_str);
		
		if (node)
		{
			node->put_text(vValue.bstrVal);
		}
		else
		{
			XML_DocumentPtr pDoc;
			pParent->get_ownerDocument(&pDoc);
			XML_NodePtr newNode = CreateDOMTextNode(pDoc, bstrChildNodeName, vValue);
			pParent->appendChild(newNode);
		}
	}
}

///////////////////////////////////////////////////////////////////////////////////
// Function: CreateDOMTextNode
///////////////////////////////////////////////////////////////////////////////////
//
// Create a text node and (converting if necessary) set the contents from parameter
// If no contents then do nothing
// Note: VT_EMPTY still creates a node (maybe it shouldn't?)
//
XML_NodePtr CreateDOMTextNode(XML_DocumentPtr pDoc, CComBSTR bstrName, CComVariant vContents)
{
	XML_NodePtr pNode;

	if (vContents.vt != VT_NULL)
	{
		pNode = CreateDOMNode(pDoc, MSXML::NODE_ELEMENT, bstrName);
		
		if (vContents.vt != VT_BSTR)
			vContents.ChangeType(VT_BSTR);
		pNode->put_text(vContents.bstrVal);
	}
	return pNode;
}

////////////////////////////////////////////////////////////////////////////////////
// Function: CreateDOMNode
////////////////////////////////////////////////////////////////////////////////////
//
// Create an empty node
//
XML_NodePtr CreateDOMNode(XML_DocumentPtr pDoc, int type, CComBSTR bstrName)
{
	XML_NodePtr pNode;

	if (pDoc.p)
	{
		CComVariant vType = (int)type;
		pNode = pDoc->createNode(vType, bstrName.m_str, L"");
	}

	return pNode;
}

XML_NodePtr CreateDOMNode(XML_DocumentPtr spDoc, const wstring& sName)
{
	XML_NodePtr spNode;
	if (spDoc.p)
	{
		CComVariant vType = static_cast<int>(MSXML::NODE_ELEMENT);
		spNode = spDoc->createNode(vType, _bstr_t(sName.c_str()), L"");
	}
	return spNode;
}


////////////////////////////////////////////////////////////////////////////////////
// AppendTextNode
////////////////////////////////////////////////////////////////////////////////////
//
// create and append text node, if there's text, or anything convertible to text, in the vContents, else
// do nothing
//
XML_NodePtr AppendTextNode(XML_DocumentPtr spDoc, XML_NodePtr spParent, CComBSTR bstrName, CComVariant vContents)
{
	XML_NodePtr spNewNode = CreateDOMTextNode(spDoc, bstrName, vContents);

	if (spNewNode.p)
	{
		spParent->appendChild(spNewNode);
	}
	return spNewNode;
}

/////////////////////////////////////////////////////////////////////////////
// append a node with a CDATA section in it
/////////////////////////////////////////////////////////////////////////////

XML_NodePtr AppendCDATANode(XML_DocumentPtr spDoc, XML_NodePtr spParent, CComBSTR bstrName, CComVariant vContents)
{
	XML_NodePtr spNew = AppendNode(spDoc, spParent, bstrName);
	spNew->appendChild(spDoc->createCDATASection(VarToText(vContents).m_str));
	return spNew;
}


////////////////////////////////////////////////////////////////////////////////////
// AppendNode
////////////////////////////////////////////////////////////////////////////////////
//
// Creates an empty node and appends it to the parent
// Returns the appended node pointer
// Intention: use to create complex nodes holding other nodes
//
XML_NodePtr AppendNode(XML_DocumentPtr spDoc, XML_NodePtr spParent, CComBSTR bstrName)
{
	XML_NodePtr spNewNode = CreateDOMNode(spDoc, MSXML::NODE_ELEMENT, bstrName.m_str);
	if (spNewNode.p)
	{
		spParent->appendChild(spNewNode);
	}
	return spNewNode;
}

////////////////////////////////////////////////////////////////////////////////////
// VarToText
////////////////////////////////////////////////////////////////////////////////////
//
// depending on content of Field, make a decent text expression
//
CComBSTR VarToText(const VARIANT& var)
{
	CComBSTR	bstr;

	CComVariant v(var);

	if (v.vt != VT_NULL && v.vt != VT_EMPTY)
	{
		if (v.vt != VT_BSTR)
		{
			v.ChangeType(VT_BSTR);
		}
		bstr = v.bstrVal;
	}
	return bstr;
}

////////////////////////////////////////////////////////////////////////////////////
// VarToLong
////////////////////////////////////////////////////////////////////////////////////
//
// try to make a long
// if not, make it zero
//
long VarToLong(const VARIANT& var)
{
	long lRetVal(0);

	CComVariant v(var);
	if (v.vt != VT_NULL && v.vt != VT_EMPTY)
	{
		if (v.vt != VT_I4) v.ChangeType(VT_I4);
		if (v.vt != VT_ERROR) lRetVal = v.lVal;
	}
	return lRetVal;
}

/////////////////////////////////////////////////////////////////////////////
// VarToDouble
// try make a double, if not possible, make a 0.0
/////////////////////////////////////////////////////////////////////////////
double VarToDouble(const VARIANT& var)
{
	double dblRetVal(0.0);

	CComVariant v(var);
	if (v.vt != VT_NULL && v.vt != VT_EMPTY)
	{
		v.ChangeType(VT_R8);
		if (v.vt == VT_R8) 
			dblRetVal = v.dblVal;
	}
	return dblRetVal;
}


};	// namespace m