// ==============================================================
// Name:			x2c - filelex.cpp
//
// Description:		input file lexer
//
// Copyright (C) 2001-2008, J. Martin Wehlou, Man in The Middle AB
// martin@wehlou.com, martin@mitm.se, MiTM AB, Flintstensvagen 10, 75267 Uppsala, Sweden
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// $LastChangedDate: $
// $LastChangedRevision: $
// $LastChangedBy: $
// $HeadURL: $
// $Id: $
// ==============================================================


#include "stdafx.h"
#include "string.h"
#pragma warning(disable:4786)
#include <sstream>
#include "exceptions.h"
#include "lex.h"
#include "constants.h"
#include "filelex.h"

using std::wstringstream;
using std::wcout;
using std::endl;


// ------------------------------------------------------------------------- //

//CFileLex::CFileLex(ifstream& is) :
//m_is(is)
CFileLex::CFileLex(const wstring& sInputFile)
{
	wifstream is;
	long lsl = openInputFile(sInputFile, &is);	
	loadFile(is);
	closeInputFile(&is, lsl);
	restart();

//	if (k_bDebug)
//		dumpvi();
}

CFileLex::~CFileLex()
{}

// ------------------------------------------------------------------------- //

long CFileLex::openInputFile(const wstring& sFileName, wifstream* pis)
{
	pis->open(sFileName.c_str(), std::ios::binary);
	if (!(*pis))
		throw x::file_error(x::v2s(L"could not open template file: %s", sFileName.c_str()));
	m_srcloc.push_back(srcloc(m_vi.size(), sFileName));
	return m_srcloc.size() - 1;
	return 0;
}

// ------------------------------------------------------------------------- //

void CFileLex::closeInputFile(wifstream* pis, long lsl)
{
	pis->close();
	m_srcloc[lsl].m_lLast = m_vi.size() - 1;
}

// ------------------------------------------------------------------------- //

void CFileLex::loadFile(wifstream& is)
{
	wchar_t buf[1024];
	wstring sIncludeName;
	long lBufTotal = 0;
	while(true)
	{
		is.getline(buf, sizeof(buf)/sizeof(buf[0]));
		if (!is) break;
		long lLen = wcslen(buf);
		if (lLen > 0 && (buf[lLen-1] == 0x0A || buf[lLen-1] == 0x0D))		// is this right for a wide string?
			buf[lLen-1] = L'\0';

		// even it it's an "#include" it gets added to the input. The included file
		// is inserted after the "#include". The parser ignores all "#" lines.
		m_vi.push_back(wstring(buf));

		// 2008 ... abolished the trial/full nonsense
		//if (k_bTrialVersion)
		//{
		//	if (m_vi.size() >= k_lTrialMaxScriptLines) throw x::lexerror(L"script too large for free version");
		//	lBufTotal += wcslen(buf);
		//	if (lBufTotal > k_lTrialMaxScriptSize) throw x::lexerror(L"script too large for free version");
		//}
		if (hasInclude(buf, &sIncludeName))
		{
			wifstream nested;
			long lsl = openInputFile(sIncludeName, &nested);
			loadFile(nested);
			closeInputFile(&nested, lsl);
		}
	}
}

// ------------------------------------------------------------------------- //

bool CFileLex::hasInclude(const wchar_t* buf, wstring* psIncludeName)
{
	const wchar_t pszInclude[] = L"#include";

	// #include has to be first in the line, possibly preceded by whitespace
	while(iswspace(*buf)) buf++;
	if (wcsncmp(buf, pszInclude, sizeof(pszInclude)/sizeof(pszInclude[0]) - 1) != 0)
		return false;

	// bump past the keyword and whitespace
	buf += sizeof(pszInclude)/sizeof(pszInclude[0]);
	while(iswspace(*buf)) buf++;

	if (*buf != L'\"')
		throwLexError(L"expected filename to include");
	buf++;

	psIncludeName->erase();
	// include all characters up to closing quote
	while (*buf != 0 && *buf != L'\"')		// is this right? the zero, I mean...
		*psIncludeName += static_cast<wchar_t>(*buf++);
	if (*buf != L'\"')
		throwLexError(L"unterminated file name string in include");

	return true;
}

// ------------------------------------------------------------------------- //

void CFileLex::restart()
{
	m_buf[0] = L'\0';
	m_cp = m_buf;
	setBufPtr(m_buf);
	m_lvi = 0;
	clearLastToken();
	loadCurrentLine();
}

// ------------------------------------------------------------------------- //

void CFileLex::skipWhite()
{
	for(;;)
	{
		while(iswspace(*m_cp))
			m_cp++;
		if (*m_cp != L'\0') break;
		nextLine();
	}
}

// ------------------------------------------------------------------------- //

void CFileLex::nextLine()
{
	if (m_lvi >= m_vi.size() - 1)
		throw x::eof();
	m_lvi++;
	loadCurrentLine();
}
	
// ------------------------------------------------------------------------- //

void CFileLex::loadCurrentLine()
{
	wcsncpy(m_buf, m_vi[m_lvi].c_str(), sizeof(m_buf)/sizeof(m_buf[0]));
	m_cp = m_buf;
}

// ------------------------------------------------------------------------- //

wstring CFileLex::getPositionDescription() const
{
	// find the source file this line comes from
	long lLineNb = 0;
	wstring sSrcName;
	for (long l = m_srcloc.size() - 1; l >= 0; l--)
	{
		if (m_lvi >= m_srcloc[l].m_lFirst && m_lvi <= m_srcloc[l].m_lLast)
		{
			lLineNb = m_lvi - m_srcloc[l].m_lFirst + 1;
			sSrcName = m_srcloc[l].m_sFileName;
			break;
		}
	}

	wstringstream ss;
//	ss << "line(" << m_lvi + 1 << "), col(" << m_cp - m_buf << ")";
	ss << sSrcName << L"(" << lLineNb << L")";
	return ss.str();
}

// ------------------------------------------------------------------------- //

void CFileLex::savepos()
{
	m_vPos.push_back(make_pair(m_lvi, m_cp - m_buf));
}

// ------------------------------------------------------------------------- //

void CFileLex::restorepos()
{
	if (m_vPos.size() < 1)
		throw x::internal_error(L"lex position stack empty on restorepos");
	m_lvi = m_vPos.back().first;
	loadCurrentLine();
	m_cp = m_buf + m_vPos.back().second;

}

// ------------------------------------------------------------------------- //

void CFileLex::savepos(const wstring& sPosName)
{
	m_mNamedPos[sPosName] = make_pair(m_lvi, m_cp - m_buf);
}

// ------------------------------------------------------------------------- //

void CFileLex::restorepos(const wstring& sPosName)
{
	t_mStrPos::iterator itr = m_mNamedPos.find(sPosName);
	if (itr == m_mNamedPos.end())
		throw x::internal_error(L"restore to unknown named position");
	m_lvi = itr->second.first;
	loadCurrentLine();
	m_cp = m_buf + itr->second.second;
}


void CFileLex::disposepos()
{
	if (m_vPos.size() < 1)
		throw x::internal_error(L"lex position stack underflow");
	m_vPos.pop_back();
}

// ------------------------------------------------------------------------- //

wstring CFileLex::readToString(const wstring& s)
{
	wstring sRV;

	while(true)
	{
		const wchar_t* cp = wcsstr(m_cp, s.c_str());
		if (cp == NULL)
		{
			sRV += readToEOL();
			nextLine();
			sRV += L"\n";
		}
		else
		{
			wstring sTmp(m_cp);
			sRV += sTmp.substr(0, (cp - m_cp));
			m_cp = cp;
			return sRV;
		}
	}
}

// ==========================================================================
//  debug extras
//
//
// ==========================================================================

void CFileLex::dumpvi()
{
	if (k_bDebug)
	{
		long l;
		wcout << L"dump of m_vi array:" << endl;
		for (l = 0; l < m_vi.size(); l++)
		{
			wcout << l << L": " << m_vi[l] << endl;
		}
		wcout << endl;
		wcout << L"dump of m_srcloc array: " << endl;
		for (l = 0; l < m_srcloc.size(); l++)
		{
			wcout << m_srcloc[l].m_lFirst << L" - " << m_srcloc[l].m_lLast << L" : " << m_srcloc[l].m_sFileName << endl;
		}
	}
}

