// ==============================================================
// Name:		x2c - funcs.cpp
//
// Description: utility functions
//
// Copyright (C) 2001-2008, J. Martin Wehlou, Man in The Middle AB
// martin@wehlou.com, martin@mitm.se, MiTM AB, Flintstensvagen 10, 75267 Uppsala, Sweden
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// $LastChangedDate: $
// $LastChangedRevision: $
// $LastChangedBy: $
// $HeadURL: $
// $Id: $
// ==============================================================

#include <string>
using std::wstring;

#include "funcs.h"
#include "exceptions.h"

using m::XML_DocumentPtr;
using m::XML_ParseErrorPtr;

namespace f
{
	// ------------------------------------------------------------------------- //
	
	wstring getPath(const wstring sName)
	{
		// the path is everything up to and including the last
		// slash or colon in the string
		for (wstring::size_type i = sName.length(); i != wstring::npos; i--)
		{
			if (sName[i] == '\\' || sName[i] == ':')
				return sName.substr(0, i + 1);
		}
		return L"";
	}

	// ------------------------------------------------------------------------- //
	
	wstring addDefaultPath(const wstring sName, const wstring sDefaultPath)
	{
		if (getPath(sName).length() > 0)
			return sName;
		return sDefaultPath + sName;
	}

	// ------------------------------------------------------------------------- //
	
	wstring::size_type findExtension(const wstring sName)
	{
		// if there is an extension, then seen from the end, there's a dot
		// before any slashes or colons
		for (wstring::size_type i = sName.length(); i != wstring::npos; i--)
		{
			if (sName[i] == '.') return i;
			if (sName[i] == '\\' || sName[i] == ':') break;
		}
		return wstring::npos;
	}
	
	// ------------------------------------------------------------------------- //
	
	wstring removeExtension(const wstring sName)
	{
		wstring::size_type pos = findExtension(sName);
		if (pos != wstring::npos)
			return sName.substr(0, pos);
		else
			return sName;
	}
	
	// ------------------------------------------------------------------------- //
	
	wstring addDefaultExtension(const wstring sName, const wstring sDefaultExtension)
	{
		wstring sRV;
		wstring sLocalName;

		if (findExtension(sName) != wstring::npos)
			sLocalName = removeExtension(sName);
		else
			sLocalName = sName;
		sRV = sLocalName + L'.' + sDefaultExtension;
		return sRV;
	}
	
	// ------------------------------------------------------------------------- //
	
	wstring replaceExtension(const wstring sName, const wstring sNewExtension)
	{
		wstring sLocalName = removeExtension(sName);
		wstring sRV = sLocalName + L'.' + sNewExtension;
		return sRV;
	}

	//
	//	convert reserved characters to escape sequences
	//	this routine assumes no characters have been escaped yet
	//
	wstring xmlEscape(const wstring& sInput)
	{
		wstring sRV;
		for (int i = 0; i < sInput.length(); ++i)
		{
			switch(sInput[i])
			{
			case L'&':
				sRV += L"&amp;";
				break;

			case L'>':
				sRV += L"&gt;";
				break;

			case L'<':
				sRV += L"&lt;";
				break;

			case L'\"':
				sRV += L"&quote;";
				break;

			case L'\'':
				sRV += L"&apos;";
				break;

			default:
				sRV += sInput[i];
				break;
			}
		}
		return sRV;
	}		

	// ------------------------------------------------------------------------- //

	void makeDirectory(const wstring& sDirName)
	{
		wchar_t buf[1024];
		bool bSuccess = true;

		GetCurrentDirectory(sizeof(buf)/sizeof(buf[0]), buf);
		if (!SetCurrentDirectory(sDirName.c_str()))
		{
			if (!CreateDirectory(sDirName.c_str(), NULL))
				bSuccess = false;
		}
		SetCurrentDirectory(buf);
		if (!bSuccess)
			throw x::file_error(x::v2s(L"could not create directory: %s", sDirName.c_str()));
	}

	// ------------------------------------------------------------------------- //

	void	openPackageStream(ofstream* os, const wstring& sPackageDir,
								const wstring& sPackageName, const wstring& sObjectName)
	{
		wstring sDirName = pathConcat(sPackageDir, sPackageName);

		try
		{
			makeDirectory(sPackageDir);
		}
		catch(const x::file_error&) {}

		try
		{
			makeDirectory(sDirName);
		}
		catch (const x::file_error&) {}
			// ignore not being able to create (it already exists, probably)

		wstring sOsName = pathConcat(sPackageDir, sPackageName) ;
		sOsName = pathConcat(sOsName, sObjectName + L".xml");
		os->open(sOsName.c_str());
	}

	// ------------------------------------------------------------------------- //

	// concatenate two strings with one backslash between them
	wstring	pathConcat(const wstring& lhs, const wstring& rhs)
	{
		wstring sRV = lhs;
		if (sRV.length() > 0 && *(sRV.rbegin()) != L'\\') 
			sRV += L"\\";
		sRV += rhs;
		return sRV;
	}

	// ------------------------------------------------------------------------- //

	XML_DocumentPtr openXmlDoc(const wstring& sFileName)
	{
		XML_DocumentPtr spDoc;
		spDoc.CoCreateInstance(__uuidof(MSXML::DOMDocument));
		if (VARIANT_TRUE != spDoc->load(_bstr_t(sFileName.c_str())))
		{
			XML_ParseErrorPtr spParseError = spDoc->GetparseError();
			throw x::xml_error(spParseError);
		}
		return spDoc;
	}

	// ------------------------------------------------------------------------- //

	wstring	stripQuotes(const wstring& sInput)
	{
		if (sInput[0] == L'\"' || sInput[0] == L'\'')
		{
			long lLen = sInput.length();
			if (lLen > 1)
			{
				wchar_t qc = sInput[0];
				if (sInput[lLen-1] == qc)
				{
					return sInput.substr(1, lLen - 2);	
				}
			}
		}
		return sInput;
	}

	// ------------------------------------------------------------------------- //
	
	wstring getlastsyserr()
	{
		DWORD dwError = GetLastError();
		LPWSTR pBuf;
		FormatMessage(
			FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
			NULL,					
			dwError,
			0,
			reinterpret_cast<LPWSTR>(&pBuf),
			0,
			NULL);
		wstring sMsg(pBuf);
		LocalFree(pBuf);
		return sMsg;
	}

	// ------------------------------------------------------------------------- //
	
	wstring getlastsyserr(const wstring& sFormat)
	{
		wstring sErr = getlastsyserr();
		wstring sMsg = v2s(sFormat.c_str(), sErr.c_str());
		return sMsg;
	}

	// ------------------------------------------------------------------------- //
	
	wstring v2s(const wchar_t* pszFormat, ...) {
		va_list		args;
		wchar_t		szBuf[512];
		int			nBuf;
		va_start	(args, pszFormat);
		nBuf = _vsnwprintf(szBuf, sizeof(szBuf)/sizeof(szBuf[0]), pszFormat, args);
		va_end(args);
		return wstring(szBuf);
	}

	
};