// ==============================================================
// Name:		x2c - lex.cpp
//
// Description:	file lexer / tokenizer
//
// Copyright (C) 2001-2008, J. Martin Wehlou, Man in The Middle AB
// martin@wehlou.com, martin@mitm.se, MiTM AB, Flintstensvagen 10, 75267 Uppsala, Sweden
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// $LastChangedDate: $
// $LastChangedRevision: $
// $LastChangedBy: $
// $HeadURL: $
// $Id: $
// ==============================================================


#include "stdafx.h"
#include <math.h>
#include "exceptions.h"
#include "lex.h"


// ==========================================================================
//  CTkn implementation
//
//	publics
// ==========================================================================

CTkn::~CTkn() 
{
	delete m_ts;
}

// ==========================================================================
//  CLex implementation
//
//	publics
// ==========================================================================


CLex::CLex() 
{
	loadTokenMap();
	m_tLast = tNONE;
}

CLex::~CLex()
{}


// ------------------------------------------------------------------------- //

void CLex::loadTokenMap()
{
	m_tokenmap[L"codepage"]	= tCODEPAGE;
	m_tokenmap[L"function"]	= tFUNCTION;
	m_tokenmap[L"nodelist"]	= tDECLNODELIST;
	m_tokenmap[L"default"]	= tDEFAULT;
	m_tokenmap[L"foreach"]	= tFOREACH;
//	m_tokenmap[L"include"]	= tINCLUDE;
	m_tokenmap[L"program"]	= tPROGRAM;
	m_tokenmap[L"append"]	= tAPPEND;
	m_tokenmap[L"create"]	= tCREATE;
	m_tokenmap[L"define"]	= tDEFINE;
	m_tokenmap[L"double"]	= tDECLDOUBLE;
	m_tokenmap[L"ifndef"]	= tIFNDEF;
	m_tokenmap[L"output"]	= tOUTPUT;
	m_tokenmap[L"return"]	= tRETURN;
	m_tokenmap[L"string"]	= tDECLSTRING;
	m_tokenmap[L"break"]	= tBREAK;
	m_tokenmap[L"close"]	= tCLOSE;
	m_tokenmap[L"endif"]	= tENDIF;
	m_tokenmap[L"error"]	= tERROR;
	m_tokenmap[L"false"]	= tFALSE;
	m_tokenmap[L"ifdef"]	= tIFDEF;
	m_tokenmap[L"where"]	= tWHERE;
	m_tokenmap[L"while"]	= tWHILE;
	m_tokenmap[L"bool"]		= tDECLBOOL;
	m_tokenmap[L"copy"]		= tCOPY;
	m_tokenmap[L"else"]		= tELSE;
	m_tokenmap[L"file"]		= tFILE;
	m_tokenmap[L"node"]		= tDECLNODE;
	m_tokenmap[L"open"]		= tOPEN;
	m_tokenmap[L"read"]		= tREAD;
	m_tokenmap[L"true"]		= tTRUE;
	m_tokenmap[L"for"]		= tFOR;
	m_tokenmap[L"map"]		= tDECLMAP;
	m_tokenmap[L"set"]		= tDECLSET;
	m_tokenmap[L"xml"]		= tXML;
	m_tokenmap[L"as"]		= tAS;
	m_tokenmap[L"if"]		= tIF;
	m_tokenmap[L"in"]		= tIN;
}

// ------------------------------------------------------------------------- //

token CLex::peekToken()
{
	if (m_tLast == tNONE)
		scanToken();
	return m_tLast;
}

// ------------------------------------------------------------------------- //

token CLex::lookupKeyword(const wstring& s)
{
	t_mStrTok::iterator itr = m_tokenmap.find(s);
	if (itr != m_tokenmap.end())
		return itr->second;
	else
		return tNONE;
}

// ------------------------------------------------------------------------- //
// return the whole token for cpstore

CTkn CLex::readCTkn()
{
	CTkn ctkn = scanCTkn();
	if (ctkn.getToken() != tEOF)
		advanceLastToken();
	return ctkn;
}


CTkn CLex::scanCTkn()
{
	for(;;)
	{
		try
		{
			skipWhite();
		}
		catch(const x::eof&)
		{
			return CTkn(tEOF);
		}
		
		if (isalpha(*m_cp))
		{
			scanAlpha(&m_cpEnd, &m_sLast);
			m_tLast = lookupKeyword(m_sLast);
			if (m_tLast == tNONE)
				m_tLast = tVARNAME;
			return CTkn(m_tLast, m_sLast);
		}
		
		else if (isdigit(*m_cp))
		{
			scanNumber(&m_cpEnd, &m_dblLast);
			m_tLast = tNUMBER;
			return CTkn(m_tLast, m_dblLast);
		}
		
		else if (*m_cp == '\'' || *m_cp == '\"')
		{
			scanQString(&m_cpEnd, &m_sLast);
			m_tLast = tQSTRING;
			return CTkn(m_tLast, m_sLast);
		}
		
		else
		{
			m_tLast = scanOperator(&m_cpEnd);
			if (m_tLast == tDSLASH)
				skipToEOL();
			else if (m_tLast == tBEGCMT)
				skipPastToken(tENDCMT);
			else
				return CTkn(m_tLast);
		}
	}
}



// ------------------------------------------------------------------------- //

// start scanning at current position and return pointer
// to position past end of found token
// if token is operator, return token
// if token is keyword, return token and string value
// if token is variable name or label, return token and string value
// if token is number, return token and numerical value
// if token is quoted string, return token and string contents


void CLex::scanToken()
{
	for(;;)
	{
		try
		{
			skipWhite();
		}
		catch(const x::eof&)
		{
			m_tLast = tEOF;
			return;
		}
		
		if (isalpha(*m_cp))
		{
			scanAlpha(&m_cpEnd, &m_sLast);
			m_tLast = lookupKeyword(m_sLast);
			if (m_tLast == tNONE)
				m_tLast = tVARNAME;
			break;
		}
		
		else if (isdigit(*m_cp))
		{
			scanNumber(&m_cpEnd, &m_dblLast);
			m_tLast = tNUMBER;
			break;
		}
		
		else if (*m_cp == '\'' || *m_cp == '\"')
		{
			scanQString(&m_cpEnd, &m_sLast);
			m_tLast = tQSTRING;
			break;
		}
		
		else
		{
			m_tLast = scanOperator(&m_cpEnd);
			if (m_tLast == tDSLASH)
				skipToEOL();
			else if (m_tLast == tBEGCMT)
				skipPastToken(tENDCMT);
			else
				break;
		}
	}
}

// ----------------------------------------------------------------- //

void CLex::scanAlpha(const wchar_t** cpEnd, wstring* psLastToken)
{
	const wchar_t* cp = m_cp;
	psLastToken->erase();
	while (isalnum(*cp) || *cp == '_')
		*psLastToken += *cp++;
	*cpEnd = cp;
}

// ---------------------------------------------------------------- //

void CLex::scanNumber(const wchar_t** cpEnd, double* pdbl)
{
	const wchar_t *cp = m_cp;
	double dblNumber = 0.0;

	bool bNeg = false;
	if (*cp == '+')
		cp++;
	else if (*cp == '-')
	{
		cp++;
		bNeg = true;
	}
	if (!isdigit(*cp))
		throw x::wrong_token();

	while (isdigit(*cp))
	{
		dblNumber *= 10;
		dblNumber += (*cp) - '0';
		cp++;
	}

	if (*cp == '.')
	{
		cp++;
		double dblDiv = 1;
		while (isdigit(*cp))
		{
			dblDiv /= 10;
			dblNumber += ((*cp) - '0') * dblDiv;
			cp++;
		}
	}

	if (*cp == 'e')
	{
		cp++;
		int iExp = 0;
		while (isdigit(*cp))
		{
			iExp *= 10;
			iExp += (*cp) - '0';
			cp++;
		}
		dblNumber = pow(dblNumber, iExp);
	}

	if (bNeg) dblNumber = -dblNumber;

	*cpEnd = cp;
	*pdbl = dblNumber;
}

// ------------------------------------------------------------------- //

void CLex::scanQString(const wchar_t** cpEnd, wstring* ps)
{
	const wchar_t* cp = m_cp;
	ps->erase();
	wchar_t qc;

	if (*cp != L'\'' && *cp != L'\"')
		throw x::wrong_token();

	qc = *cp;
	cp++;

	while (*cp != L'\0' && *cp != qc)
		*ps += *cp++;

	if (*cp == qc)
		cp++;			// bump past ending quote
	else
		throw x::lexerror(L"unterminated string");

	*cpEnd = cp;
}

// ------------------------------------------------------------------------- //

wstring CLex::readVarName()
{
	if (m_tLast == tNONE) 
		scanToken();
	wstring s = m_sLast;
	if (0 == s.length())
		throw x::lexerror(L"missing variable name");
	advanceLastToken();
	return s;
}

// ------------------------------------------------------------------------- //

wstring CLex::readString()
{
	if (peekToken() != tQSTRING)
		throwLexError(L"expected a quoted string");
	wstring s = m_sLast;
	advanceLastToken();
	return s;
}

// ------------------------------------------------------------------------- //

double CLex::readNumber()
{
	if (peekToken() != tNUMBER)
		throwLexError(L"expected a number");
	double dbl = m_dblLast;
	advanceLastToken();
	return dbl;
}

// ------------------------------------------------------------------------- //

token CLex::readToken()
{
	token t = peekToken();
	advanceLastToken();
	return t;
}

// ------------------------------------------------------------------------- //

bool CLex::ifToken(token t)
{
	if (t == peekToken())
	{
		advanceLastToken();
		return true;
	}
	else
		return false;
}

// ------------------------------------------------------------------------- //

token CLex::matchToken(token t)
{

	if (peekToken() != t)
	{
		switch (t) {
			case tSEMICOLON:
				throw x::lexerror(L"missing semicolon");
			default:
				throw x::lexerror(L"missing token on match");
		}
	}
	advanceLastToken();
	return t;
}

// ==========================================================================
//  CLex implementation
//
//	protected
// ==========================================================================

void CLex::setBufPtr(const wchar_t* pbuf)
{
	m_pBuf = pbuf;
	m_cp = pbuf;
}

// ------------------------------------------------------------------------- //

token CLex::scanOperator(const wchar_t** pcp)
{
	enum eState {
		sSTART, sALPHA, sNUMBER,
		sCOMMA, sDOT,
		sSLASH, sSLASH_SLASH, sSLASH_STAR, sSLASH_EQ,
		sAND, sAND_AND, sAND_EQ,
		sPIPE, sPIPE_PIPE,
		sLT, sLT_EQ,
		sGT, sGT_EQ,
		sEQ, sEQ_EQ,
		sEXC, sEXC_EQ,
		sLSQ, sLSQ_LSQ,
		sRSQ, sRSQ_RSQ,
		sSTAR, sSTAR_SLASH, sSTAR_EQ,
		sBSL,
		sCOL,
		sSCOL,
		sLPAR,
		sRPAR,
		sLBRC,
		sRBRC,
		sPLUS, sPLUS_PLUS, sPLUS_EQ,
		sMIN, sMIN_MIN, sMIN_EQ,
		sHASH,
	};

	const wchar_t* cp = m_cp;
	*pcp = cp;

	int state = sSTART;

	for(;;)
	{
		switch(state)
		{
		case sSTART:		switch(*cp)
							{
							case ',':	state = sCOMMA;	break;
							case '.':	state = sDOT;	break;
							case '/':	state = sSLASH;	break;
							case '&':	state = sAND;	break;
							case '|':	state = sPIPE;	break;
							case '<':	state = sLT;	break;
							case '>':	state = sGT;	break;
							case '=':	state = sEQ;	break;
							case '!':	state = sEXC;	break;
							case '[':	state = sLSQ;	break;
							case ']':	state = sRSQ;	break;
							case '*':	state = sSTAR;	break;
							case '\\':	state = sBSL;	break;
							case ':':	state = sCOL;	break;
							case ';':	state = sSCOL;	break;
							case '(':	state = sLPAR;	break;
							case ')':	state = sRPAR;	break;
							case '{':	state = sLBRC;	break;
							case '}':	state = sRBRC;	break;
							case '+':	state = sPLUS;	break;
							case '-':	state = sMIN;	break;
							case '#':	state = sHASH;	break;
							default:	return tNONE;
							}
							break;

		case sCOMMA:				return tCOMMA;
		case sDOT:					return tDOT;

		case sSLASH:		if (*cp == '/')			state = sSLASH_SLASH; 
							else if (*cp == '*')	state = sSLASH_STAR; 
							else if (*cp == '=')	state = sSLASH_EQ; 
							else	return tSLASH;
							break;
		case sSLASH_SLASH:			return tDSLASH;
		case sSLASH_STAR:			return tBEGCMT;
		case sSLASH_EQ:				return tDIVASS;

		case sAND:			if (*cp == '&')			state = sAND_AND;
							else if (*cp == '=')	state = sAND_EQ; 
							else	return tAMPERSAND;
							break;
		case sAND_AND:				return tAND;
		case sAND_EQ:				return tCONCATASS;


		case sPIPE:			if (*cp == '|')			state = sPIPE_PIPE;
							else	return tPIPE;
							break;
		case sPIPE_PIPE:			return tOR;

		case sLT:			if (*cp == '=')			state = sLT_EQ;
							else	return tLT;
							break;
		case sLT_EQ:				return tLE;

		case sGT:			if (*cp == '=')			state = sGT_EQ;
							else	return tGT;
							break;
		case sGT_EQ:				return tGE;

		case sEQ:			if (*cp == '=')			state = sEQ_EQ;
							else	return tASSIGN;
							break;
		case sEQ_EQ:				return tEQ;

		case sEXC:			if (*cp == '=')			state = sEXC_EQ;
							else	return tNOTOP;
							break;
		case sEXC_EQ:				return tNE;

		case sLSQ:			if (*cp == '[')			state = sLSQ_LSQ;
							else	throw x::wrong_token();
							break;
		case sLSQ_LSQ:				return tOPENDSQB;

		case sRSQ:			if (*cp == ']')			state = sRSQ_RSQ;
							else	throw x::wrong_token();
							break;
		case sRSQ_RSQ:				return tCLOSEDSQB;

		case sSTAR:			if (*cp == '/')			state = sSTAR_SLASH;
							else if (*cp == '=')	state = sSTAR_EQ;
							else	return tMULTIPLY;
							break;
		case sSTAR_SLASH:			return tENDCMT;
		case sSTAR_EQ:				return tMULTASS;

		case sBSL:					return tBACKSLASH;
		case sCOL:					return tCOLON;
		case sSCOL:					return tSEMICOLON;
		case sLPAR:					return tOPENPAREN;
		case sRPAR:					return tCLOSEPAREN;
		case sLBRC:					return tOPENBRACE;
		case sRBRC:					return tCLOSEBRACE;

		case sPLUS:			if (*cp == '+')			state = sPLUS_PLUS;
							else if (*cp == '=')	state = sPLUS_EQ;
							else	return tPLUS;
							break;
		case sPLUS_PLUS:			return tINCREMENT;
		case sPLUS_EQ:				return tPLUSASS;

		case sMIN:			if (*cp == '-')			state = sMIN_MIN;
							else if (*cp == '=')	state = sMIN_EQ;
							else	return tMINUS;
							break;
		case sMIN_MIN:				return tDECREMENT;
		case sMIN_EQ:				return tMINASS;
		case sHASH:					return tHASH;
		}

		cp++;
		*pcp = cp;
	}
}

// ------------------------------------------------------------------------- //

wstring CLex::readToEOL()
{
	wstring s;
	while (*m_cp != L'\0')
	{
		s += *m_cp;
		++m_cp;
	}
	clearLastToken();
	return s;
}

// ------------------------------------------------------------------------- //

void CLex::skipToEOL()
{
	while (*m_cp != L'\0')
		m_cp++;
	clearLastToken();
}

// ------------------------------------------------------------------------- //

void CLex::skipPastToken(token t)
{
	while (t != readToken());
}

// ------------------------------------------------------------------------- //

void CLex::skipToToken(token t)
{
	while (t != peekToken())
		readToken();
}

// ------------------------------------------------------------------------- //

void CLex::advanceLastToken()
{
	if (m_tLast == tNONE)
		throwLexError(L"last token unknown");

	m_cp = m_cpEnd;
	clearLastToken();
}

// ------------------------------------------------------------------------- //

void CLex::clearLastToken()
{
	m_tLast = tNONE;
	m_sLast = L"";
}

// ------------------------------------------------------------------------- //

void CLex::throwLexError(const wstring& sMsg)
{
	throw x::lexerror(sMsg);
}

// ------------------------------------------------------------------------- //

wstring CLex::getErrorInfo() const
{
	wstring sErrInfo;
	sErrInfo += getPositionDescription() + L"\n";
	sErrInfo += m_pBuf;
	sErrInfo += L'\n';

	for (const wchar_t* cp = m_pBuf; cp != m_cp; cp++)
	{
		if (*cp == L'\t')
			sErrInfo += L'\t';
		else
			sErrInfo += L' ';
	}

	sErrInfo += L'^';					// uparrow marker
	sErrInfo += L'\n';
	return sErrInfo;
}

// ------------------------------------------------------------------------- //
// static help functions

eValType CLex::token2vt(token t)
{
	switch(t)
	{
	case tDECLSTRING:	return vtSTR;
	case tDECLDOUBLE:	return vtDBL;
	case tDECLBOOL:		return vtBOOL;
	case tDECLNODE:		return vtNODE;
	case tDECLNODELIST:	return vtNODELIST;
	case tDECLMAP:		return vtMAP;	
	case tDECLSET:		return vtSET;	
	default:			return vtNONE;
	}
}