// ==============================================================
// Name:		x2c - parse.cpp
//
// Description:	the actual real syntax parser for x2c
//				it's a recursive decent parser, parsing the source text
//				directly without first preparing a syntax tree (yeah, I know,
//				that would have been both neater and better, but remember,
//				I was young then, relatively speaking)
//
// Copyright (C) 2001-2008, J. Martin Wehlou, Man in The Middle AB
// martin@wehlou.com, martin@mitm.se, MiTM AB, Flintstensvagen 10, 75267 Uppsala, Sweden
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// $LastChangedDate: $
// $LastChangedRevision: $
// $LastChangedBy: $
// $HeadURL: $
// $Id: $
// ==============================================================

#include "stdafx.h"
#pragma warning(disable:4786)
#include <iostream>
#include "exceptions.h"
#include "lex.h"
#include "stringlex.h"
#include "MSXMLImport.h"
#include "valstack.h"
#include "builtins.h"
#include "userfunc.h"
#include "vset.h"
#include "vstr.h"
#include "vxml.h"
#include "parse.h"


using std::wcout;
using std::endl;
using std::wofstream;

ofstream	CParse::m_os;
CSymbols*	CParse::m_pRootSym = NULL;

// ------------------------------------------------------------------------- //

CParse::CParse() : m_pSym(NULL), m_bBreak(false), m_iBreakable(0)
{}

CParse::~CParse()
{
	if (m_bIsRootParser) {
		delete m_pSym;
	}
}

// ------------------------------------------------------------------------- //

void CParse::run(CLex* pLex)
{
	try
	{
		m_pSym = new CSymbols();
		m_pRootSym = m_pSym;
		m_bIsRootParser = true;
		CBuiltins::loadsymbols(*m_pSym);
		buildFuncTbl(pLex);

		CValStack valstack;
		doStatements(pLex, true, &valstack);
	}
	catch(const x::symbol_error& e)
	{
		SyntaxError(pLex, e.msg());
	}
	catch(const x::syntaxerror& e)
	{
		SyntaxError(pLex, e.msg());
	}
	catch(const x::wrong_token&)
	{
		SyntaxError(pLex, L"unexpected token");
	}
	catch(const _com_error& e)
	{
		SyntaxError(pLex, e.ErrorMessage());
	}
	catch(const x::xml_error& e)
	{
		SyntaxError(pLex, e.msg());
	}
	catch(const x::lexerror& e)
	{
		SyntaxError(pLex, e.msg());
	}
	catch(const x::stackunderflow&)
	{
		SyntaxError(pLex, L"stack underflow, missing parameter?");
	}
	catch(...)
	{
		SyntaxError(pLex, L"unhandled exception");
	}
}

// ------------------------------------------------------------------------- //

void CParse::runsubparser(CLex* pLex, CParse* pRootParser, CSymbols* pSym, CValStack* pvs,
						  CFunctorBase* pCurrentFunc)
{
	m_pSym = pSym;
	m_bIsRootParser = false;
	m_pRootParser = pRootParser;
	m_pCurrentFunc = pCurrentFunc;

	// subparser always runs a block of {...} something
	// but a return can make you miss the ending brace
	pLex->matchToken(tOPENBRACE);
	doStatements(pLex, true, pvs);
	pLex->ifToken(tCLOSEBRACE);
}

// ------------------------------------------------------------------------- //

// scan the whole input and load function addresses
// (this is where I should have built a syntax tree instead)

void CParse::buildFuncTbl(CLex* pLex)
{
	CValStack valstack;			// a dummy, really
	for(;;)
	{
		token t = pLex->peekToken();

		switch(t)
		{
		case tFUNCTION:
			addFuncTbl(pLex);
			break;

		case tEOF:
			pLex->restart();
			return;

		default:
			if (!doStatement(pLex, false, &valstack))
				pLex->matchToken(t);
			break;
		}

	}
}

// ------------------------------------------------------------------------- //
//	function fname(<paramlist>)
//	function fname(<paramlist>) : returntype
//  note that return type (if any) need to be declared

void CParse::addFuncTbl(CLex* pLex)
{
	pLex->matchToken(tFUNCTION);
	wstring sFuncName = pLex->readVarName();
	pLex->matchToken(tOPENPAREN);

	t_vpVtStr vpVtStr;
	eValType vtReturns = vtNONE;

	if (pLex->peekToken() != tCLOSEPAREN)
	{
		for(;;)
		{
			token tType = pLex->readToken();
			eValType vtType = pLex->token2vt(tType);
			wstring sParamName = pLex->readVarName();
			vpVtStr.push_back(make_pair(vtType, sParamName));
			if (!pLex->ifToken(tCOMMA)) break;
		}
	}
	pLex->matchToken(tCLOSEPAREN);

	// check for return value spec
	if (pLex->ifToken(tCOLON))
	{
		token tRetType = pLex->readToken();
		vtReturns = pLex->token2vt(tRetType);
		if (vtReturns == vtNONE)	
			throw x::syntaxerror(L"unknown return type");
	}

	CUserFunc* pUserFunc = new CUserFunc(sFuncName, vtReturns, vpVtStr, pLex);
	CFunctorBaseRcp rcpUserFunc(pUserFunc);

	m_pSym->defineFunc(sFuncName, rcpUserFunc);
	pLex->savepos(sFuncName);		
}

// ------------------------------------------------------------------------- //

void CParse::doStatements(CLex* pLex, bool bEnable, CValStack* pvs)
{
	for (;;)
	{
		// once break is set, ignore the rest of the statement block
		// (yet another thing that would have been much better if I had
		// used a syntax tree model)
		if (!doStatement(pLex, bEnable && !m_bBreak, pvs)) break;
	}
}

// ------------------------------------------------------------------------- //

bool CParse::doStatement(CLex* pLex, bool bEnable, CValStack* pvs)
{
	token t = pLex->peekToken();
	
	switch(t)
	{
	case tFOREACH:		doForeach(pLex, bEnable, pvs);
						break;
		
	case tIF:			doIf(pLex, bEnable, pvs);
						break;
		
	case tFOR:			doFor(pLex, bEnable, pvs);
						break;
		
	case tWHILE:		doWhile(pLex, bEnable, pvs);
						break;
		
	case tDEFINE:		doDefine(pLex, bEnable, pvs);
						break;
		
	case tIFDEF:		doIfdef(pLex, bEnable, pvs);
						break;
		
	case tIFNDEF:		doIfndef(pLex, bEnable, pvs);
						break;
		
	case tOPEN:			doOpen(pLex, bEnable, pvs);
						pLex->matchToken(tSEMICOLON);
						break;
		
	case tCLOSE:		doClose(pLex, bEnable, pvs);
						pLex->matchToken(tSEMICOLON);
						break;
		
	case tOUTPUT:		doOutput(pLex, bEnable, pvs);
						pLex->matchToken(tSEMICOLON);
						break;
		
	case tDECLDOUBLE:	doDeclDouble(pLex, bEnable, pvs);
						pLex->matchToken(tSEMICOLON);
						break;
		
	case tDECLSTRING:	doDeclString(pLex, bEnable, pvs);
						pLex->matchToken(tSEMICOLON);
						break;
		
	case tDECLNODE:		doDeclNode(pLex, bEnable, pvs);
						pLex->matchToken(tSEMICOLON);
						break;
		
	case tDECLNODELIST:	doDeclNodelist(pLex, bEnable, pvs);
						pLex->matchToken(tSEMICOLON);
						break;
		
	case tDECLMAP:		doDeclMap(pLex, bEnable, pvs);
						pLex->matchToken(tSEMICOLON);
						break;
		
	case tDECLSET:		doDeclSet(pLex, bEnable, pvs);
						pLex->matchToken(tSEMICOLON);
						break;
		
	case tVARNAME:		doVarname(pLex, bEnable, pvs);
						pLex->matchToken(tSEMICOLON);
						break;
		
	case tOPENBRACE:	// statement list between { and }
						m_pSym->addScope();
						pLex->matchToken(tOPENBRACE);
						doStatements(pLex, bEnable, pvs);
						pLex->matchToken(tCLOSEBRACE);
						m_pSym->dropScope();
						break;
		
	case tOPENDSQB:		// output to file (between double square brackets "[[ ... ]]")
						doDBSQ(pLex, bEnable, pvs);
						break;
		
	case tDOT:			doDotLine(pLex, bEnable, pvs);
						break;
		
	case tRETURN:		doReturn(pLex, bEnable, pvs);
						if (bEnable) return false;
						break;
		
	case tBREAK:		doBreak(pLex, bEnable, pvs);
						break;
		
	case tSEMICOLON:	pLex->matchToken(tSEMICOLON);
						break;
		
	case tHASH:			pLex->skipToEOL();
						break;
		
	case tFUNCTION:		return false;
		
	case tEOF:			return false;

	case tCLOSEBRACE:	return false;

	default:			throw x::syntaxerror(L"Unexpected token");
	}

	return true;
}

// ==========================================================================
//  statement handlers
//
//
// ==========================================================================

// foreach string f in <set>
// foreach node n in <nodelist>
void CParse::doForeach(CLex* pLex, bool bEnable, CValStack* pvs)
{
	m_iBreakable++;
	m_pSym->addScope();
	pLex->matchToken(tFOREACH);
	
	if (pLex->ifToken(tDECLSTRING))
		doForeachString(pLex, bEnable, pvs);
	
	else if (pLex->ifToken(tDECLNODE))
		doForeachNode(pLex, bEnable, pvs);
	
	else
		throw x::wrong_token();
	
	m_pSym->dropScope();
	m_iBreakable--;
}

// ------------------------------------------------------------------------- //

void CParse::doForeachString(CLex* pLex, bool bEnable, CValStack* pvs)
{
	if (bEnable)
	{
		wstring sStringVar = pLex->readVarName();
		m_pSym->defineStr(sStringVar);
		pLex->matchToken(tIN);
		wstring sListVar = pLex->readVarName();

		eValType vt = m_pSym->getValType(sListVar);
		if (vt == vtSET)
		{
			CVSet& rSET = m_pSym->getVSet(sListVar);

			if (rSET.begin() == rSET.end())
				doStatement(pLex, false, pvs);
			else
			{
				pLex->savepos();
				for (CVSet::iterator itr = rSET.begin(); itr != rSET.end(); ++itr)
				{
					pLex->restorepos();
					m_pSym->getVStr(sStringVar) = *itr;
					m_pSym->addScope();
					doStatement(pLex, true, pvs);
					m_pSym->dropScope();
					if (m_bBreak) 
					{
						m_bBreak = false;
						break;
					}
				}
				pLex->disposepos();
			}
		}
	}
	else	// not enabled
	{
		pLex->readVarName();
		pLex->matchToken(tIN);
		pLex->readVarName();
		doStatement(pLex, false, pvs);
	}
}

// ------------------------------------------------------------------------- //

void CParse::doForeachNode(CLex* pLex, bool bEnable, CValStack* pvs)
{
	if (bEnable)
	{
		wstring sNodeVar = pLex->readVarName();
		m_pSym->defineNode(sNodeVar);
		pLex->matchToken(tIN);
		genexpr(pLex, pvs);
		CNodelist& rNL = pvs->popNodelist();
		
		// if nothing there, scan past the statement(s)
		if (rNL.begin() == rNL.end())
			doStatement(pLex, false, pvs);
		
		else
		{
			pLex->savepos();
			for (CNodelist::iterator itr = rNL.begin(); itr != rNL.end(); ++itr)
			{
				pLex->restorepos();
				m_pSym->getNodeVal(sNodeVar) = *itr;
				m_pSym->addScope();
				doStatement(pLex, true, pvs);
				m_pSym->dropScope();
				if (m_bBreak) 
				{
					m_bBreak = false;
					break;
				}
			}
			pLex->disposepos();
		}
	}
	else	// not enabled
	{
		pLex->readVarName();
		pLex->matchToken(tIN);
		skipgenexpr(pLex);
		doStatement(pLex, false, pvs);
	}
}

// ------------------------------------------------------------------------- //

void CParse::doIf(CLex* pLex, bool bEnable, CValStack* pvs)
{
	pLex->matchToken(tIF);
	bool bRes = false;
	if (bEnable)
	{
		pLex->matchToken(tOPENPAREN);
		genexpr(pLex, pvs);
		pLex->matchToken(tCLOSEPAREN);
		bRes = pvs->popBool();
	}
	else
		skipParenExpr(pLex, pvs);
	
	doStatement(pLex, bEnable && bRes, pvs);
	if (pLex->ifToken(tELSE))
		doStatement(pLex, bEnable && !bRes, pvs);
}

// ------------------------------------------------------------------------- //

void CParse::skipParenExpr(CLex* pLex, CValStack* pvs)
{
	pLex->matchToken(tOPENPAREN);

	long lLevel = 0;
	for(;;)
	{
		token t = pLex->readToken();
		switch(t)
		{
		case tOPENPAREN:
			lLevel++;
			break;

		case tCLOSEPAREN:
			if (lLevel == 0) return;
			lLevel--;
			break;

		case tDOT:
			doDotLine(pLex, false, pvs);
			break;

		case tOPENDSQB:
			doDBSQ(pLex, false, pvs);
			break;

		default:
			break;
		}
	}
}

// ------------------------------------------------------------------------- //

void CParse::doFor(CLex* pLex, bool bEnable, CValStack* pvs)
{
	m_iBreakable++;
	m_pSym->addScope();
	pLex->matchToken(tFOR);
	m_pSym->dropScope();
	m_iBreakable--;
}

// ------------------------------------------------------------------------- //

void CParse::doWhile(CLex* pLex, bool bEnable, CValStack* pvs)
{
	m_iBreakable++;
	if (bEnable)
	{
		m_pSym->addScope();
		pLex->matchToken(tWHILE);
		pLex->savepos();
		while(true)
		{
			pLex->restorepos();
			genexpr(pLex, pvs);
			bool bEval = pvs->popBool();
			doStatement(pLex, bEval, pvs);
			if (!bEval) break;
			if (m_bBreak) 
			{
				m_bBreak = false;
				break;
			}
		}
		pLex->disposepos();
		m_pSym->dropScope();
	}
	else
	{
		pLex->matchToken(tWHILE);
		genexpr(pLex, pvs);
		pvs->popBool();
		doStatement(pLex, false, pvs);
	}
	m_iBreakable--;
}

// ------------------------------------------------------------------------- //

void CParse::doDefine(CLex* pLex, bool bEnable, CValStack* pvs)
{
	pLex->matchToken(tDEFINE);
}

// ------------------------------------------------------------------------- //

void CParse::doIfdef(CLex* pLex, bool bEnable, CValStack* pvs)
{
	pLex->matchToken(tIFDEF);
}

// ------------------------------------------------------------------------- //

void CParse::doIfndef(CLex* pLex, bool bEnable, CValStack* pvs)
{
	pLex->matchToken(tIFNDEF);
}

// ------------------------------------------------------------------------- //

// open xml filename as xf;
// open create filename;
// open append filename;
// open <c/a> filename as unicode;		(default)
// open <c/a> filename as codepage <nb>;	(write converting to MCBS using given codepage)
// (open create xml filename as xf;)
void CParse::doOpen(CLex* pLex, bool bEnable, CValStack* pvs)
{
	if (bEnable)
	{
		pLex->matchToken(tOPEN);
		
		if (pLex->ifToken(tXML))
		{
			genexpr(pLex, pvs);
			wstring sFileName = pvs->popStr();
			pLex->matchToken(tAS);
			wstring sFileLabel = pLex->readVarName();
			m_pSym->defineXMLRef(sFileLabel, sFileName);
		}
		else if (pLex->ifToken(tCREATE) || pLex->ifToken(tAPPEND))
		{
			std::ios::open_mode om = std::ios::out | std::ios::binary | ((pLex->ifToken(tCREATE)) ? std::ios::trunc : std::ios::app);

			genexpr(pLex, pvs);
			wstring sFileName = pvs->popStr();
			ofstream& os = (m_bIsRootParser) ? m_os : m_pRootParser->getOs();

			if (os.is_open())
				os.close();
			os.open(sFileName.c_str(), std::ios::out | std::ios::trunc | std::ios::binary);
			if (!os)
				SyntaxError(pLex, L"could not create output file");


			if (pLex->ifToken(tAS)) {
				UINT codepage = 0;
				if (pLex->ifToken(tUNICODE)) {
					codepage = 0;
				}
				else if (pLex->ifToken(tCODEPAGE)) {
					genexpr(pLex, pvs);
					codepage = int(pvs->popDbl());
				}
				else
					SyntaxError(pLex, L"Expected 'unicode' or 'codepage' here");

				if (m_bIsRootParser) {
					m_codepage = codepage;
				}
				else {
					m_pRootParser->setCodepage(codepage);
				}
			}

			// if we're doing unicode, the file needs a byte order header
			if (0 == m_codepage) {
				wchar_t BOM = 0xFEFF;
				os.write((char*)&BOM, sizeof(wchar_t));
			}
		}
		else
			throw x::wrong_token();
	}
	else
		pLex->skipToToken(tSEMICOLON);
}

// ------------------------------------------------------------------------- //

// close xf;
void CParse::doClose(CLex* pLex, bool bEnable, CValStack* pvs)
{
	pLex->matchToken(tCLOSE);
	wstring sFileLabel = pLex->readVarName();
	if (bEnable)
		m_pSym->undefine(sFileLabel);
}

// ------------------------------------------------------------------------- //

void CParse::doOutput(CLex* pLex, bool bEnable, CValStack* pvs)
{
	pLex->matchToken(tOUTPUT);
}

// ------------------------------------------------------------------------- //

// double varname;
// double varname = expression;
void CParse::doDeclDouble(CLex* pLex, bool bEnable, CValStack* pvs)
{
	if (bEnable)
	{
		pLex->matchToken(tDECLDOUBLE);
		wstring sVarName = pLex->readVarName();

		if (pLex->ifToken(tASSIGN))
		{
			genexpr(pLex, pvs);
			double dblValue = pvs->popDbl();
			m_pSym->defineDbl(sVarName, dblValue);
		}
		else
			m_pSym->defineDbl(sVarName);
	}
	else
		pLex->skipToToken(tSEMICOLON);
}

// ------------------------------------------------------------------------- //

void CParse::doDeclString(CLex* pLex, bool bEnable, CValStack* pvs)
{
	if (bEnable)
	{
		pLex->matchToken(tDECLSTRING);
		wstring sVarName = pLex->readVarName();
			
		if (pLex->ifToken(tASSIGN))
		{
			genexpr(pLex, pvs);
			wstring sValue = pvs->popStr();
			m_pSym->defineStr(sVarName, sValue);
		}
		else
			m_pSym->defineStr(sVarName);
	}
	else
		pLex->skipToToken(tSEMICOLON);
}

// ------------------------------------------------------------------------- //

void CParse::doDeclBool(CLex* pLex, bool bEnable, CValStack* pvs)
{
	if (bEnable)
	{
		pLex->matchToken(tDECLBOOL);
		wstring sVarName = pLex->readVarName();

		if (pLex->ifToken(tASSIGN))
		{
			genexpr(pLex, pvs);
			bool bValue = pvs->popBool();
			m_pSym->defineBool(sVarName, bValue);
		}
		else
			m_pSym->defineBool(sVarName, false);
	}
	else
		pLex->skipToToken(tSEMICOLON);
}

// ------------------------------------------------------------------------- //

void CParse::doDeclNode(CLex* pLex, bool bEnable, CValStack* pvs)
{
	if (bEnable)
	{
		pLex->matchToken(tDECLNODE);
		wstring sVarName = pLex->readVarName();

		if (pLex->ifToken(tASSIGN))
		{
			genexpr(pLex, pvs);
			XML_NodePtr spNode = pvs->popNode();
			m_pSym->defineNode(sVarName, spNode);
		}
		else 
			m_pSym->defineNode(sVarName);
	}
	else
		pLex->skipToToken(tSEMICOLON);
}

// ------------------------------------------------------------------------- //

void CParse::doDeclNodelist(CLex* pLex, bool bEnable, CValStack* pvs)
{
	if (bEnable)
	{
		pLex->matchToken(tDECLNODELIST);
		wstring sVarName = pLex->readVarName();
		CNodelist& rNL = m_pSym->defineNodelist(sVarName);
			
		if (pLex->ifToken(tASSIGN))
		{
			genexpr(pLex, pvs);
			rNL = pvs->popNodelist();
		}
	}
	else
		pLex->skipToToken(tSEMICOLON);
}

// ------------------------------------------------------------------------- //

// map <name>;
// map <name> = { <key> : <value>,..., default : <defvalue> };
// <key> = <string> | <number>;
// <value> = <string> | <number>;
// <defvalue> = <string> | <number> | error | copy
void CParse::doDeclMap(CLex* pLex, bool bEnable, CValStack* pvs)
{
	if (bEnable)
	{
		pLex->matchToken(tDECLMAP);
		wstring sVarName = pLex->readVarName();
		CVMap& rM = m_pSym->defineMap(sVarName);
		if (pLex->ifToken(tASSIGN))
		{
			mapExpr(pLex, bEnable, pvs);
			rM = pvs->popVMap();
		}
	}
	else
		pLex->skipToToken(tSEMICOLON);
}

// ------------------------------------------------------------------------- //

// set <name>;
// set <name> = {<value>,...,<value>};
// <value> = <string>
void CParse::doDeclSet(CLex* pLex, bool bEnable, CValStack* pvs)
{
	if (bEnable)
	{
		pLex->matchToken(tDECLSET);
		wstring sVarName = pLex->readVarName();
		CVSet& rS = m_pSym->defineSet(sVarName);
		if (pLex->ifToken(tASSIGN))
		{
			setExpr(pLex, bEnable, pvs);
			rS = pvs->popVSet();
		}
	}
	else
		pLex->skipToToken(tSEMICOLON);
}

// ------------------------------------------------------------------------- //

void CParse::doVarname(CLex* pLex, bool bEnable, CValStack* pvs)
{
	// varname++;
	// varname--;
	// varname = expr;
	// varname(params);
	
	if (bEnable)
	{
		wstring sVarName = pLex->readVarName();
		eValType vt = m_pSym->getValType(sVarName);
		
		if (vt == vtFUNC)
		{
			// using a discardable stack so any ignored return
			// values won't screw things up
			CValStack valstack;
			func(pLex, sVarName, &valstack);
		}
		else
		{
			token t = pLex->peekToken();
			switch(t)
			{
			case tINCREMENT:
				pLex->matchToken(t);
				++(m_pSym->getDbl(sVarName));
				break;
				
			case tDECREMENT:
				pLex->matchToken(t);
				--(m_pSym->getDbl(sVarName));
				break;
				
			case tASSIGN:
				pLex->matchToken(t);
				assignTo(pLex, sVarName, bEnable, pvs, vt);
				break;

			case tPLUSASS:
			case tMINASS:
			case tMULTASS:
			case tDIVASS:
			case tCONCATASS:
				pLex->matchToken(t);
				opAssignTo(pLex, sVarName, bEnable, pvs, vt, t);
				break;
				
			default:
				SyntaxError(pLex, L"unexpected token");
				break;
			}
		}
	}
	else
		pLex->skipToToken(tSEMICOLON);
}

// ------------------------------------------------------------------------- //

void CParse::doDBSQ(CLex* pLex, bool bEnable, CValStack* pvs)
{
	pLex->matchToken(tOPENDSQB);
	// retrieve multiline until end of output
	wstring s = pLex->readToString(L"]]");	
	if (bEnable)
	{
		s = processLine(pLex, s, pvs);
		outputLine(pLex, s);
	}
	pLex->matchToken(tCLOSEDSQB);
}

// ------------------------------------------------------------------------- //

void CParse::doDotLine(CLex* pLex, bool bEnable, CValStack* pvs)
{
	pLex->matchToken(tDOT);
	wstring s = pLex->readToEOL();
	if (bEnable)
	{
		s = processLine(pLex, s, pvs);
		s += L"\n";
		outputLine(pLex, s);
	}
}

// ------------------------------------------------------------------------- //

void CParse::doReturn(CLex* pLex, bool bEnable, CValStack* pvs)
{
	pLex->matchToken(tRETURN);
	if (bEnable)
	{
		token t = pLex->peekToken();
		eValType vtReturns = vtNONE;
		if (!m_bIsRootParser)
			vtReturns = m_pCurrentFunc->getReturnType();

		if (t == tSEMICOLON)
		{
			if (!m_bIsRootParser && (vtReturns != vtNONE))
				throw x::syntaxerror(L"must return a value");
		}
		else
		{
			if (!m_bIsRootParser && (vtReturns == vtNONE))
				throw x::syntaxerror(L"must not return a value");
			genexpr(pLex, pvs);
		}

		pLex->matchToken(tSEMICOLON);
	}
	else
		pLex->skipToToken(tSEMICOLON);
}

// ------------------------------------------------------------------------- //

// breaks out of the current statement block (!)
void CParse::doBreak(CLex* pLex, bool bEnable, CValStack* pvs)
{
	pLex->matchToken(tBREAK);
	if (m_iBreakable < 1)
		SyntaxError(pLex, L"break without a foreach, for, or while");
	if (bEnable) m_bBreak = true;
}

// ==========================================================================
//	end of statement implementations
// ==========================================================================

void CParse::pushVar(CLex* pLex, const wstring& sVarName, CValStack* pvs)
{
	eValType vt = m_pSym->getValType(sVarName);
	switch(vt)
	{
	case vtSTR:
		pvs->push(m_pSym->getVStr(sVarName).get());
		break;

	case vtDBL:
		pvs->push(m_pSym->getDbl(sVarName));
		break;

	case vtNODE:
		pvs->push(m_pSym->getNodeVal(sVarName).getNode());
		break;

	case vtNODELIST:
		pvs->push(m_pSym->getNodelist(sVarName));
		break;
	}
}


// ------------------------------------------------------------------------- //

wstring CParse::processLine(CLex* pLex, const wstring& sOutputLine, CValStack* pvs)
{
	wstring s = sOutputLine;
	while(true)
	{
		// replace back to front
		wstring::size_type bpos = s.rfind(L"<$");
		if (bpos == wstring::npos) break;
		wstring::size_type epos = s.find(L">", bpos);
		if (epos == wstring::npos)
			SyntaxError(pLex, L"unterminated embedded expression");

		// now bpos and epos point like this:
		// ...yadda yadda <$yadda> yadda ...
		//           bpos ^      ^ epos
		// so [bpos+2, epos-1] is the expression to evaluate and has
		// length = epos - bpos - 2, while
		// replacing [bpos,epos]

		wstring sExpr = s.substr(bpos + 2, epos - bpos - 2);
		CStringLex strLex(sExpr);
		genexpr(&strLex, pvs);
		wstring sEval = pvs->popStr();
		s.replace(bpos, epos - bpos + 1, sEval);
	}
	return s;
}

// ------------------------------------------------------------------------- //

void CParse::outputLine(CLex* pLex, const wstring& sOutputLine)
{
	ofstream& os = (m_bIsRootParser) ? m_os : m_pRootParser->getOs();
	UINT codepage = (m_bIsRootParser) ? m_codepage : m_pRootParser->getCodepage();
	if (!os.is_open())
		SyntaxError(pLex, L"writing to closed output stream");

	// handle unicode if that's what we're doing
	if (0 == m_codepage) {
		os.write((char*)sOutputLine.c_str(), sOutputLine.length() * sizeof(wchar_t));
		if (os.fail()) {
			throw x::internal_error(L"output stream failed");
		}
	}
	else {
		UINT bufsize = WideCharToMultiByte(codepage, 0, sOutputLine.c_str(), sOutputLine.length(), NULL, 0, NULL, NULL);
		char* outBuf = new char[bufsize];
		BOOL usedDefaultChar = false;
		int result = WideCharToMultiByte(codepage, 0, sOutputLine.c_str(), sOutputLine.length(), outBuf, bufsize, NULL, &usedDefaultChar);
		if (0 == result) {
			DWORD le = GetLastError();
			if (ERROR_INSUFFICIENT_BUFFER == le)
				throw x::internal_error(L"insufficient buffer in outputLine for codepage conversion");
			if (ERROR_INVALID_FLAGS == le)
				throw x::internal_error(L"invalid flags in outputLine for codepage conversion");
			if (ERROR_INVALID_PARAMETER == le) {
				wstringstream wss;
				wss << L"you probably specified a code page (" << m_codepage << ") that is not installed on this system";
				throw x::syntaxerror(wss.str());
			}
			throw x::internal_error(L"unknown error occurred while converting output line to specified codepage");
		}
		else {
			if (usedDefaultChar) {
				wcout << endl;
				wcout << L"something in the following text could not be converted to given codepage:\n";
				wcout << L"--- start of problematic text ---" << endl;
				wcout << sOutputLine << endl;
				wcout << L"--- end of problematic text ---" << endl << endl;
			}
			os.write(outBuf, bufsize);
			if (os.fail()) {
				throw x::internal_error(L"output stream failed");
			}
		}
	}

}


// ------------------------------------------------------------------------- //

void CParse::SyntaxError(CLex* pLex, const wstring& sMsg)
{
	wstring s = L"Syntax error:";
	s += sMsg;
	s += L'\n';
	s += pLex->getErrorInfo();
	wcout << s << endl;
}



// ------------------------------------------------------------------------- //
// { <key> : <value>, .... default : <defvalue> }
// or function or variable reference

void CParse::mapExpr(CLex* pLex, bool bEnable, CValStack* pvs)
{
	CVMap mv;
	if (bEnable)
	{
		if (pLex->ifToken(tOPENBRACE))
		{
			// braced constant expression
			while(true)
			{
				token t = pLex->peekToken();
				
				switch (t)
				{
				case tQSTRING:
					{
						wstring sKey = pLex->readString();
						pLex->matchToken(tCOLON);
						wstring sValue = pLex->readString();
						mv.add(sKey, sValue);
					}
					break;
					
				case tDEFAULT:
					{
						pLex->matchToken(tDEFAULT);
						pLex->matchToken(tCOLON);
						t = pLex->peekToken();
						if (t == tQSTRING)
						{
							wstring sDefVal = pLex->readString();
							mv.setDef(sDefVal);
						}
						else if (t == tERROR)
						{
							pLex->matchToken(tERROR);
							mv.setErrDefault();
						}
						else if (t == tCOPY)
						{
							pLex->matchToken(tCOPY);
							mv.setCopyDefault();
						}
						else
							SyntaxError(pLex, L"illegal default value in map");
						
					}
					break;
					
				default:
					break;
				}
				
				if (!pLex->ifToken(tCOMMA)) break;
			}
			pLex->matchToken(tCLOSEBRACE);
		}
		else
			// any other expression (like: map m = x;)
			genexpr(pLex, pvs);
	}
	else
		pLex->skipToToken(tSEMICOLON);
		
	pvs->push(mv);
}

// ------------------------------------------------------------------------- //
// { string, string,...,string }
void CParse::setExpr(CLex* pLex, bool bEnable, CValStack* pvs)
{
	if (bEnable)
	{
		if (pLex->ifToken(tOPENBRACE))
		{
			CVSet vSet;
			
			while(true)
			{
				genexpr(pLex, pvs);
				vSet.insert(pvs->popStr());
				if (!pLex->ifToken(tCOMMA)) break;
			}
			
			pLex->matchToken(tCLOSEBRACE);
			pvs->push(vSet);
		}
		else
			genexpr(pLex, pvs);
	}
	else
		pLex->skipToToken(tSEMICOLON);
}

// ------------------------------------------------------------------------- //

void CParse::assignTo(CLex* pLex, const wstring& sVarName, bool bEnable, CValStack* pvs, eValType vt)
{
	if (bEnable)
	{
		switch(vt)
		{
		case vtBOOL:
			genexpr(pLex, pvs);
			m_pSym->getBool(sVarName) = pvs->popBool();
			break;

		case vtDBL:
			genexpr(pLex, pvs);
			m_pSym->getDbl(sVarName) = pvs->popDbl();
			break;
			
		case vtSTR:
			genexpr(pLex, pvs);
			m_pSym->getVStr(sVarName) = pvs->popStr();
			break;
			
		case vtNODE:
			genexpr(pLex, pvs);
			m_pSym->getNodeVal(sVarName) = pvs->popNode();
			break;

		case vtNODELIST:
			genexpr(pLex, pvs);
			m_pSym->getNodelist(sVarName) = pvs->popNodelist();
			break;
			
		case vtMAP:
			if (pLex->peekToken() == tOPENBRACE)
				mapExpr(pLex, bEnable, pvs);
			else
				genexpr(pLex, pvs);
			m_pSym->getVMap(sVarName) = pvs->popVMap();
			break;

		case vtSET:
			if (pLex->peekToken() == tOPENBRACE)
				setExpr(pLex, bEnable, pvs);
			else
				genexpr(pLex, pvs);
			m_pSym->getVSet(sVarName) = pvs->popVSet();
			break;
		}
	}
	else
		pLex->skipPastToken(tSEMICOLON);
}

// ------------------------------------------------------------------------- //

void CParse::opAssignTo(CLex* pLex, const wstring& sVarName, bool bEnable, CValStack* pvs, eValType vt, token t)
{
	if (bEnable)
	{
		switch(vt)
		{
		case vtBOOL:
			SyntaxError(pLex, L"operator not defined for bool");
			break;

		case vtDBL:
			switch(t)
			{
			case tPLUSASS:
				genexpr(pLex, pvs);
				m_pSym->getDbl(sVarName) += pvs->popDbl();
				break;

			case tMINASS:
				genexpr(pLex, pvs);
				m_pSym->getDbl(sVarName) -= pvs->popDbl();
				break;

			case tMULTASS:
				genexpr(pLex, pvs);
				m_pSym->getDbl(sVarName) *= pvs->popDbl();
				break;

			case tDIVASS:
				genexpr(pLex, pvs);
				m_pSym->getDbl(sVarName) /= pvs->popDbl();
				break;
			
			default:
				SyntaxError(pLex, L"operator not defined for double");
				break;
			}
			break;

		case vtSTR:
			if (t == tCONCATASS)
			{
				genexpr(pLex, pvs);
				m_pSym->getVStr(sVarName).get() += pvs->popStr();
			}
			else
				SyntaxError(pLex, L"operator not defined for string");
			break;

		case vtNODE:
			SyntaxError(pLex, L"operator not defined for node");
			break;

		case vtNODELIST:
			SyntaxError(pLex, L"operator not defined for nodelist");
			break;

		case vtMAP:
			SyntaxError(pLex, L"operator not defined for map");
			break;

		case vtSET:
			if (t == tPLUSASS)
			{
				genexpr(pLex, pvs);
				m_pSym->getVSet(sVarName).insert(pvs->popStr());
			}
			else if (t == tMINASS)
			{
				genexpr(pLex, pvs);
				m_pSym->getVSet(sVarName).erase(pvs->popStr());
			}
			else
				SyntaxError(pLex, L"operator not defined for set");
			break;

		default:
			break;
		}
	}
	else
		pLex->skipPastToken(tSEMICOLON);
}

// ------------------------------------------------------------------------- //
// V 3


//	<genexpr>	::=		<expr> | <genexpr> && <expr> | <genexpr> || <expr>
//	<expr>		::=		<relterm> | <relterm> <relop> <relterm> 
//	<relterm>	::=		true | false | <subexpr>
//	<subexpr>	::=		<sign> <term> | <subexpr> & <term> | <subexpr> + <term> | <subexpr> - <term>
//	<term>		::=		<factor> | <term> * <factor> | <term> / <factor>
//	<factor>	::=		<constant> | <variable> | <func> | (<genexpr>)
//	<func>		::=		<funcname> ( <paramlist> )
//	<paramlist>	::=		<param> | <param> , <paramlist>
//	<param>		::=		<genexpr>
//
//	<relop>		::=		== | != | <= | >= | < | >
//	<sign>		::=		- | +

void CParse::genexpr(CLex* pLex, CValStack* pvs)
{
	expr(pLex, pvs);
	for(;;)
	{
		token t = pLex->peekToken();
		if (t == tAND)
		{
			pLex->matchToken(t);
			expr(pLex, pvs);
			pvs->push(pvs->popBool() && pvs->popBool());
		}
		else if (t == tOR)
		{
			pLex->matchToken(t);
			expr(pLex, pvs);
			pvs->push(pvs->popBool() || pvs->popBool());
		}
		else
			break;
	}
}

// ------------------------------------------------------------------------- //

void CParse::expr(CLex* pLex, CValStack* pvs)
{
	relterm(pLex, pvs);
	for(;;)
	{
		token t = pLex->peekToken();
		if (isRelop(t))
		{
			pLex->matchToken(t);
			relterm(pLex, pvs);
			// let the right hand side type determine comparison type
			if (pvs->isBool())
				evalboolexpr(pLex, t, pvs);
			else if (pvs->isDbl())
				evaldblexpr(pLex, t, pvs);
			else if (pvs->isStr())
				evalstrexpr(pLex, t, pvs);
			else if (pvs->isNode())
				evalnodeexpr(pLex, t, pvs);
			else if (pvs->isNodelist())
				evalnodelistexpr(pLex, t, pvs);
			else
				throw x::internal_error(L"unknown eval expression type");
		}
		else
			break;
	}
}

// ------------------------------------------------------------------------- //

void CParse::evalboolexpr(CLex* pLex, token t, CValStack* pvs)
{
	bool rhs = pvs->popBool();
	bool lhs = pvs->popBool();

	switch(t)
	{
	case tEQ:	pvs->push(rhs == lhs);	break;
	case tNE:	pvs->push(rhs != lhs);	break;
	default:
		SyntaxError(pLex, L"illegal comparison of boolean terms");
		break;
	}
}

// ------------------------------------------------------------------------- //

void CParse::evaldblexpr(CLex* pLex, token t, CValStack* pvs)
{
	double rhs = pvs->popDbl();
	double lhs = pvs->popDbl();
	switch(t)
	{
	case tEQ:	pvs->push(lhs == rhs);		break;
	case tNE:	pvs->push(lhs != rhs);		break;
	case tLT:	pvs->push(lhs < rhs);		break;
	case tGT:	pvs->push(lhs > rhs);		break;
	case tLE:	pvs->push(lhs <= rhs);		break;
	case tGE:	pvs->push(lhs >= rhs);		break;
	default:
		SyntaxError(pLex, L"illegal comparison of numeric terms");
		break;
	}
}

// ------------------------------------------------------------------------- //

void CParse::evalstrexpr(CLex* pLex, token t, CValStack* pvs)
{
	wstring rhs = pvs->popStr();
	wstring lhs = pvs->popStr();
	switch(t)
	{
	case tEQ:	pvs->push(lhs == rhs);		break;
	case tNE:	pvs->push(lhs != rhs);		break;
	case tLT:	pvs->push(lhs < rhs);		break;
	case tGT:	pvs->push(lhs > rhs);		break;
	case tLE:	pvs->push(lhs <= rhs);		break;
	case tGE:	pvs->push(lhs >= rhs);		break;
	default:
		SyntaxError(pLex, L"illegal comparison of string terms");
		break;
	}
}

// ------------------------------------------------------------------------- //

void CParse::evalnodeexpr(CLex* pLex, token t, CValStack* pvs)
{
	XML_NodePtr rhs = pvs->popNode();
	XML_NodePtr lhs = pvs->popNode();
	switch(t)
	{
	case tEQ:	pvs->push(lhs == rhs);		break;
	case tNE:	pvs->push(lhs != rhs);		break;
	default:
		SyntaxError(pLex, L"illegal comparison of nodes");
	}
}

// ------------------------------------------------------------------------- //

void CParse::evalnodelistexpr(CLex* pLex, token, CValStack*)
{
	SyntaxError(pLex, L"illegal comparison of nodelists");
}

// ------------------------------------------------------------------------- //
//	<relterm>	::=		true | false | <subexpr>

void CParse::relterm(CLex* pLex, CValStack* pvs)
{
	token t = pLex->peekToken();
	if (t == tTRUE)
	{
		pLex->matchToken(t);
		pvs->push(true);
	}
	else if (t == tFALSE)
	{
		pLex->matchToken(t);
		pvs->push(false);
	}
	else
		subexpr(pLex, pvs);
}

// ------------------------------------------------------------------------- //
//	<subexpr>	::=		<sign> <term> | <subexpr> & <term> | <subexpr> + <term> | <subexpr> - <term>

void CParse::subexpr(CLex* pLex, CValStack* pvs)
{
	term(pLex, pvs);
	for(;;)
	{
		token t = pLex->peekToken();
		if (t == tAMPERSAND)
		{
			pLex->matchToken(t);
			term(pLex, pvs);
			wstring rhs = pvs->popStr();
			wstring lhs = pvs->popStr();
			pvs->push(lhs + rhs);
		}
		else if (t == tPLUS)
		{
			pLex->matchToken(t);
			eValType leftVt = pvs->getValType();
			term(pLex, pvs);
			if (leftVt == vtDBL)
			{
				double rhs = pvs->popDbl();
				double lhs = pvs->popDbl();
				pvs->push(lhs + rhs);
			}
			else if (leftVt == vtSET)
			{
				eValType rightVt = pvs->getValType();
				if (rightVt == vtSTR)
				{
					wstring sStr = pvs->popStr();
					CVSet leftSet = pvs->popVSet();
					leftSet.insert(sStr);
					pvs->push(leftSet);
				}
				else if (rightVt == vtSET)
				{
					CVSet rightSet = pvs->popVSet();
					CVSet leftSet = pvs->popVSet();
					leftSet.insert(rightSet);
					pvs->push(leftSet);
				}
				else
					SyntaxError(pLex, L"bad datatype for minus operator on set");
			}
		}
		else if (t == tMINUS)
		{
			pLex->matchToken(t);
			eValType leftVt = pvs->getValType();
			term(pLex, pvs);
			if (leftVt == vtDBL)
			{
				double rhs = pvs->popDbl();
				double lhs = pvs->popDbl();
				pvs->push(lhs - rhs);
			}
			else if (leftVt == vtSET)
			{
				eValType rightVt = pvs->getValType();
				if (rightVt == vtSTR)
				{
					wstring sStr = pvs->popStr();
					CVSet leftSet = pvs->popVSet();
					leftSet.erase(sStr);
					pvs->push(leftSet);
				}
				else if (rightVt == vtSET)
				{
					CVSet rightSet = pvs->popVSet();
					CVSet leftSet = pvs->popVSet();
					leftSet.erase(rightSet);
					pvs->push(leftSet);
				}
				else
					SyntaxError(pLex, L"bad datatype for minus operator on set");
			}
		}
		else
			break;
	}
}

// ------------------------------------------------------------------------- //
//	<term>		::=		<factor> | <term> * <factor> | <term> / <factor>

void CParse::term(CLex* pLex, CValStack* pvs)
{
	bool bMinus = false;
	if (pLex->ifToken(tMINUS))
		bMinus = true;

	factor(pLex, pvs);
	for(;;)
	{
		token t = pLex->peekToken();
		if (t == tMULTIPLY)
		{
			pLex->matchToken(t);
			factor(pLex, pvs);
			double rhs = pvs->popDbl();
			double lhs = pvs->popDbl();
			pvs->push(lhs * rhs);
		}
		else if (t == tDIVIDE)
		{
			pLex->matchToken(t);
			factor(pLex, pvs);
			double rhs = pvs->popDbl();
			double lhs = pvs->popDbl();
			pvs->push(lhs / rhs);
		}
		else 
			break;
	}

	if (bMinus)
		pvs->push(-pvs->popDbl());
}

// ------------------------------------------------------------------------- //
//	<factor>	::=		<constant> | <variable> | <func> | (<genexpr>)

void CParse::factor(CLex* pLex, CValStack* pvs)
{
	token t = pLex->peekToken();
	if (t == tQSTRING)
	{
		wstring s = pLex->readString();
		pvs->push(s);
	}
	else if (t == tNUMBER)
	{
		double dbl = pLex->readNumber();
		pvs->push(dbl);
	}
	else if (t == tINCREMENT)
	{
		// pre-increment
		pLex->matchToken(tINCREMENT);
		wstring sVarName = factorVarName(pLex, pvs);
		double& dbl = m_pSym->getDbl(sVarName);
		dbl++;
		pvs->popDbl();
		pvs->push(dbl);
	}

	else if (t == tDECREMENT)
	{
		// pre-decrement
		pLex->matchToken(tDECREMENT);
		wstring sVarName = factorVarName(pLex, pvs);
		double& dbl = m_pSym->getDbl(sVarName);
		dbl--;
		pvs->popDbl();
		pvs->push(dbl);
	}
	else if (t == tVARNAME)
	{
		wstring sVarName = factorVarName(pLex, pvs);
		if (pLex->ifToken(tINCREMENT))
		{
			// post-increment, don't increment value on stack, only variable
			m_pSym->getDbl(sVarName)++;
		}
		else if (pLex->ifToken(tDECREMENT))
		{
			// post-decrement, don't decrement value on stack, only variable
			m_pSym->getDbl(sVarName)--;
		}
	}
	else if (t == tOPENPAREN)
	{
		pLex->matchToken(t);
		genexpr(pLex, pvs);
		pLex->matchToken(tCLOSEPAREN);
	}
}

wstring CParse::factorVarName(CLex* pLex, CValStack* pvs)
{
	wstring sName = pLex->readVarName();
	eValType vt = m_pSym->getValType(sName);
	switch(vt)
	{
	case vtSTR:			pvs->push(m_pSym->getVStr(sName).get());			break;
	case vtDBL:			pvs->push(m_pSym->getDbl(sName));					break;
	case vtNODE:		pvs->push(m_pSym->getNodeVal(sName).getNode());		break;
	case vtNODELIST:	pvs->push(m_pSym->getNodelist(sName));				break;
	case vtXMLREF:		pvs->push(m_pSym->getVXml(sName).getRoot());		break;
	case vtSET:			pvs->push(m_pSym->getVSet(sName));					break;
	case vtMAP:			pvs->push(m_pSym->getVMap(sName));					break;
		
		// in case of a function, we really should check that it pushes one
		// and only one value on the stack. But how?
	case vtFUNC:		func(pLex, sName, pvs);								break;
	default:
		SyntaxError(pLex, L"bad variable type in expression");
		break;
	}
	return sName;
}

// ------------------------------------------------------------------------- //
//	<func>		::=		<funcname> ( <paramlist> )

void CParse::func(CLex* pLex, const wstring& sFuncName, CValStack* pvs)
{
	pLex->matchToken(tOPENPAREN);
	
	if (pLex->peekToken() != tCLOSEPAREN)
	{
		while(true)
		{
			genexpr(pLex, pvs);
			if (!pLex->ifToken(tCOMMA)) break;
		}
	}
	pLex->matchToken(tCLOSEPAREN);
	
	m_pSym->getFuncVal(sFuncName)(pvs, *m_pRootSym, (m_bIsRootParser) ? this : m_pRootParser);
}

// ------------------------------------------------------------------------- //

bool CParse::isRelop(token t) const
{
	return t == tEQ || t == tNE || t == tLE || t == tGE || t == tLT || t == tGT;
}

// ------------------------------------------------------------------------- //


// ========================================================================= //
// functions to skip through an expression in disabled state
// ========================================================================= //

void CParse::skipgenexpr(CLex* pLex)
{
	skipexpr(pLex);
	for(;;)
	{
		token t = pLex->peekToken();
		if (t == tAND)
		{
			pLex->matchToken(t);
			skipexpr(pLex);
		}
		else if (t == tOR)
		{
			pLex->matchToken(t);
			skipexpr(pLex);
		}
		else
			break;
	}
}

// ------------------------------------------------------------------------- //

void CParse::skipexpr(CLex* pLex)
{
	skiprelterm(pLex);
	for(;;)
	{
		token t = pLex->peekToken();
		if (isRelop(t))
		{
			pLex->matchToken(t);
			skiprelterm(pLex);
		}
		else
			break;
	}
}

// ------------------------------------------------------------------------- //

void CParse::skiprelterm(CLex* pLex)
{
	token t = pLex->peekToken();
	if (t == tTRUE)
	{
		pLex->matchToken(t);
	}
	else if (t == tFALSE)
	{
		pLex->matchToken(t);
	}
	else
		skipsubexpr(pLex);
}

// ------------------------------------------------------------------------- //

void CParse::skipsubexpr(CLex* pLex)
{
	skipterm(pLex);
	for(;;)
	{
		token t = pLex->peekToken();
		if (t == tAMPERSAND)
		{
			pLex->matchToken(t);
			skipterm(pLex);
		}
		else if (t == tPLUS)
		{
			pLex->matchToken(t);
			skipterm(pLex);
		}
		else if (t == tMINUS)
		{
			pLex->matchToken(t);
			skipterm(pLex);
		}
		else
			break;
	}
}

// ------------------------------------------------------------------------- //

void CParse::skipterm(CLex* pLex)
{
	pLex->ifToken(tMINUS);

	skipfactor(pLex);
	for(;;)
	{
		token t = pLex->peekToken();
		if (t == tMULTIPLY)
		{
			pLex->matchToken(t);
			skipfactor(pLex);
		}
		else if (t == tDIVIDE)
		{
			pLex->matchToken(t);
			skipfactor(pLex);
		}
		else 
			break;
	}
}

// ------------------------------------------------------------------------- //

void CParse::skipfactor(CLex* pLex)
{
	token t = pLex->peekToken();
	if (t == tQSTRING)
	{
		pLex->readString();
	}
	else if (t == tNUMBER)
	{
		pLex->readNumber();
	}
	else if (t == tINCREMENT)
	{
		// pre-increment
		pLex->matchToken(tINCREMENT);
		skipfactorVarName(pLex);
	}

	else if (t == tDECREMENT)
	{
		// pre-decrement
		pLex->matchToken(tDECREMENT);
		skipfactorVarName(pLex);
	}
	else if (t == tVARNAME)
	{
		skipfactorVarName(pLex);
		if (pLex->ifToken(tINCREMENT))
		{
			// post-increment, don't increment value on stack, only variable
		}
		else if (pLex->ifToken(tDECREMENT))
		{
			// post-decrement, don't decrement value on stack, only variable
		}
	}
	else if (t == tOPENPAREN)
	{
		pLex->matchToken(t);
		skipgenexpr(pLex);
		pLex->matchToken(tCLOSEPAREN);
	}	
}

// ------------------------------------------------------------------------- //

void CParse::skipfactorVarName(CLex* pLex)
{
	wstring sName = pLex->readVarName();
	if (pLex->peekToken() == tOPENPAREN)
	{
		CValStack vs;
		skipParenExpr(pLex, &vs);
	}
}

// ------------------------------------------------------------------------- //
//	<func>		::=		<funcname> ( <paramlist> )

void CParse::skipfunc(CLex* pLex)
{
	pLex->matchToken(tOPENPAREN);
	
	if (pLex->peekToken() != tCLOSEPAREN)
	{
		while(true)
		{
			skipgenexpr(pLex);
			if (!pLex->ifToken(tCOMMA)) break;
		}
	}
	pLex->matchToken(tCLOSEPAREN);
}

