// ==============================================================
// Name:		x2c - pstore.cpp
//
// Description:	token storage
//
// Copyright (C) 2001-2008, J. Martin Wehlou, Man in The Middle AB
// martin@wehlou.com, martin@mitm.se, MiTM AB, Flintstensvagen 10, 75267 Uppsala, Sweden
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// $LastChangedDate: $
// $LastChangedRevision: $
// $LastChangedBy: $
// $HeadURL: $
// $Id: $
// ==============================================================


#include "stdafx.h"

#include <string>
#include <iostream>
#include <fstream>
#include <vector>
#include "lex.h"
#include "exceptions.h"

using std::wstring;
using std::wistream;
using std::wostream;
using std::wifstream;
using std::wofstream;
using std::vector;

#include "pstore.h"


class CTkn_if : public CTkn_sub
{
public:
	CTkn_if(long lxThen, long lxElse = -1) :
	  m_lxThen(lxThen), m_lxElse(lxElse) {};
	long getThen() const	{ return m_lxThen; }
	long getElse() const	{ return m_lxElse; }
private:
	long	m_lxThen;
	long	m_lxElse;
};

class CTkn_for : public CTkn_sub
{
public:
	CTkn_for(long lxPastEnd) : m_lxPastEnd(lxPastEnd) {}
	long getPastEnd() const	{ return m_lxPastEnd; }
private:
	long	m_lxPastEnd;
};

class CTkn_call : public CTkn_sub
{
public:
	CTkn_call(long lxFunc) : m_lxFunc(lxFunc) {}
	long getFunc() const { return m_lxFunc; }
private:
	long	m_lxFunc;
};

class CTkn_func : public CTkn_sub
{
public:
	CTkn_func(long lxBegin, long lxEnd) : m_lxBegin(lxBegin), m_lxEnd(lxEnd) {}
	long getBegin() const { return m_lxBegin; }
	long getEnd() const { return m_lxEnd; }
private:
	long	m_lxBegin;
	long	m_lxEnd;
};

// ============================================================
// 
// each token in the storage is represented by the ptoken class
//
// ============================================================

class ptoken
{
public:
	ptoken();
	ptoken(const CTkn& ctkn);
	virtual ~ptoken();

	void save(wostream& os);
	void load(wistream& is);

private:
	CTkn	m_ctkn;
};

// ============================================================
// 
// the implementation class for the token storage
//
// ============================================================

typedef vector<ptoken> t_vPTkn;

class psimpl
{
public:
	psimpl();
	virtual ~psimpl();

	void save(wostream& os);
	void load(wistream& is);
	void load(CLex* plex);

protected:
	void add(const ptoken& ptk);
	long count() const;
	ptoken& get(long lIndex);

private:
	t_vPTkn		m_vPTkn;
};

// ============================================================
// facade class implementation
// ============================================================

CpStore::CpStore()
{
	m_pimpl = new psimpl();
}

CpStore::~CpStore()
{
	delete m_pimpl;
}

// ------------------------------------------------------------
// load the store from a source in a lexer
void CpStore::load(CLex* pLex)
{
	m_pimpl->load(pLex);
}

// ------------------------------------------------------------
// load the store from a precompiled binary
void CpStore::load(const wstring& sTokenFileName)
{
	wifstream is;
	is.open(sTokenFileName.c_str(), std::ios::binary);
	if (!is)
		throw x::file_error(L"could not open input token file");
	m_pimpl->load(is);
	is.close();
}

// ------------------------------------------------------------
// save the store to a precompiled binary
void CpStore::save(const wstring& sTokenFileName)
{
	wofstream os;
	os.open(sTokenFileName.c_str(), std::ios::binary);
	if (!os)
		throw x::file_error(L"could not open output token file");
	m_pimpl->save(os);
	os.close();
}

// ============================================================
// ptoken methods
// ============================================================

ptoken::ptoken() : m_ctkn(tNONE) {}
ptoken::ptoken(const CTkn& ctkn) : m_ctkn(ctkn) {}

ptoken::~ptoken()
{}

void ptoken::load(wistream& is)
{
	if (is.eof())
		throw x::eof();
}

void ptoken::save(wostream& os)
{}

// ============================================================
// psimpl methods
// ============================================================

psimpl::psimpl()
{}

psimpl::~psimpl()
{}

// ------------------------------------------------------------

void psimpl::load(CLex* pLex)
{
	while(true)
	{
		CTkn ctkn = pLex->readCTkn();
		if (ctkn.getToken() == tEOF) break;
		add(ptoken(ctkn));
	}
	pLex->restart();
}

// ------------------------------------------------------------

void psimpl::load(wistream& is)
{
	try
	{
		while (true)
		{
			ptoken ptk;
			ptk.load(is);
			add(ptk);
		}
	}
	catch(const x::eof&)
	{
		return;
	}
}

// ------------------------------------------------------------

void psimpl::save(wostream& os)
{
	for (long l = 0; l < count(); l++)
	{
		get(l).save(os);
	}
}

// ------------------------------------------------------------

void psimpl::add(const ptoken& ptk)
{
	m_vPTkn.push_back(ptk);
}

// ------------------------------------------------------------

ptoken& psimpl::get(long lIndex)
{
	return m_vPTkn.at(lIndex);
}

// ------------------------------------------------------------

long psimpl::count() const
{
	return m_vPTkn.size();
}

// ------------------------------------------------------------

