// ==============================================================
// Name:		x2c - rcptr.h
//
// Description:	reference counted pointer... you know, I actually published
//				this template in Windows Developer Journal. I'm so cool.
//				If you want the explanation, go:
//				http://www.wehlou.com/Code/Rcptr/index.htm
//
// Copyright (C) 2001-2008, J. Martin Wehlou, Man in The Middle AB
// martin@wehlou.com, martin@mitm.se, MiTM AB, Flintstensvagen 10, 75267 Uppsala, Sweden
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// $LastChangedDate: $
// $LastChangedRevision: $
// $LastChangedBy: $
// $HeadURL: $
// $Id: $
// ==============================================================


//////////////////////////////////////////////////////////////////
// === LISTING 1 ===
//
// --- rcptr.h ---
//
// J.M.Wehlou, 2001
//
//////////////////////////////////////////////////////////////////

#ifndef RCPTR_H
#define RCPTR_H

#include <set>
using std::set;

#include <assert.h>

template <class T> class rcPtr;
template <class T> class rcWPtr;

// -----------------------------------------------------------
// base class mixin
// derive your counted class from this one

template <class T> class rcBase
{
    friend class rcPtr<T>;
    friend class rcWPtr<T>;
    
    //
    // count of strong pointers to our class
    //
    mutable long    m_lCnt;
    
    //
    // set holding pointers to all weak pointers to our class
    //
    typedef set<rcWPtr<T>*> tdSet;
    tdSet           m_wptrset;
    
    public:
        rcBase() : m_lCnt(0) {}
};

// -----------------------------------------------------------
// weak pointer

template <class T> class rcWPtr
{
    friend class rcPtr<T>;
    T*  m_ptr;
    
    public:
        rcWPtr() : m_ptr(0) {}
        
        //
        // if you create a weak pointer from a bare pointer, 
        // you should make sure the pointed-to object has at 
        // least one strong pointer already, else
        // things tend to get dicey; the object may not have 
        // a stable existence. It would be reasonable to throw 
        // an exception where there's an assertion
        // now, but that depends on your application.
        //
        rcWPtr(T* ptr) : m_ptr(ptr)
        {
            if (m_ptr)
            {
                assert(m_ptr->m_lCnt > 0);    // see comment above
                m_ptr->m_wptrset.insert(this);
            }
        }
        
        rcWPtr(const rcWPtr<T>& src)
        {
            copyfrom(src);
        }
        
        ~rcWPtr()
        {
            release();
        }
        
        rcWPtr<T>& operator= (const rcWPtr<T>& src)
        {
            if (&src != this)
            {
                release();
                copyfrom(src);
            }
            return *this;
        }
        
        bool operator== (const rcWPtr<T>& rhs) const
        {
            return m_ptr == rhs.m_ptr;
        }
        
        T* operator->() const
        {
            return m_ptr;
        }
        
        bool isnull() const
        {
            return 0 == m_ptr;
        }
        
        void release()    
        { 
            if (m_ptr)
            {
                m_ptr->m_wptrset.erase(this);
                m_ptr = 0;
            }
        }
        
        T& operator*()                
        {
            //
            // if this assert fires, you've allocated the 
            // pointer but forgot to create an object for 
            // the pointer ("new", remember?)
            //
            assert(m_ptr);
            if (0 == m_ptr)
                throw "Null pointer";
            return *m_ptr; 
        }

		const T& operator*() const
		{
			assert(m_ptr);
			if (0 == m_ptr)
				throw "Null pointer";
			return *m_ptr;
		}

    private:
        //
        // setnull() called from last strong refcounted 
        // pointer as the object gets deleted
        //
        void setnull()
        {
            m_ptr = 0;
        }
        
        void copyfrom(const rcWPtr<T>& src)
        {
            m_ptr = src.m_ptr;
            if (m_ptr) 
            {
                m_ptr->m_wptrset.insert(this);
            }
        }
};

// -----------------------------------------------------------
// strong pointer

template <class T> class rcPtr  
{
    T*  m_ptr;
    
    public:
        rcPtr() : m_ptr(0)    {}
        
        rcPtr(T* ptr) : m_ptr(ptr)
        { 
            if (m_ptr)
            {
                m_ptr->m_lCnt++;
            }
        }
        
        rcPtr(const rcPtr<T>& src) 
        {
            copyfrom(src);
        }
        
        ~rcPtr() 
        { 
            release(); 
        }
        
        rcPtr<T>& operator= (const rcPtr<T>& src)
        {
            if (&src != this)
            {
                release();
                copyfrom(src);
            }
            return *this;
        }
        
        //
        // operator to check if two reference counting pointers 
        // point to the same underlying object
        //
        bool operator== (const rcPtr<T>& rhs) const
        {
            return m_ptr == rhs.m_ptr;
        }
        
        rcPtr<T>& attach(T* p)
        {
            release();
            m_ptr = p;
            if (m_ptr)
                m_ptr->m_lCnt++;
            return *this;
        }
        
        T* operator->() const        
        { 
            //
            // if this assert fires, you've allocated the pointer 
            // but forgot to create an object for the pointer 
            // ("new", remember?)
            //
            assert(m_ptr);
            return m_ptr; 
        }
        
        rcWPtr<T> getwptr() const
        {
            assert(0 != m_ptr);
            assert(0 < m_ptr->m_lCnt);
            return rcWPtr<T>(m_ptr);
        }
        
        bool isnull() const            
        { 
            return m_ptr == NULL; 
        }
        
        //
        // The release() method needs to set all still active
        // weak pointers to null if the strong reference
        // count goes to zero. This is the main difference
        // with the other templates.
        //
        void release()    
        { 
            if (m_ptr)
            {
                m_ptr->m_lCnt--;
                if (m_ptr->m_lCnt == 0)
                {
                    //
                    // tell all weak ptrs we're dying
                    //
                    rcBase<T>::tdSet::const_iterator cit;
                    for (cit = m_ptr->m_wptrset.begin(); 
                    cit != m_ptr->m_wptrset.end(); 
                    ++cit)
                    {
                        (*cit)->setnull();
                    }
                    delete m_ptr;
                }
                m_ptr = 0;
            }
        }
        
        T& operator*()                
        {
            //
            // if this assert fires, you've allocated the pointer 
            // but forgot to create an object for the pointer 
            // ("new", remember?)
            //
            assert(m_ptr);
            if (0 == m_ptr)
                throw "Null pointer";
            return *m_ptr; 
        }

		const T& operator*() const
		{
			assert(m_ptr);
			if (0 == m_ptr)
				throw "Null pointer";
			return *m_ptr;
		}
        
    private:
        void copyfrom(const rcPtr<T>& src)
        {
            m_ptr = src.m_ptr;
            if (m_ptr) 
            {
                m_ptr->m_lCnt++;
            }
        }
    };

#endif