// ==============================================================
// Name:		x2c - symbols.cpp
//
// Description:	symbol table
//
// Copyright (C) 2001-2008, J. Martin Wehlou, Man in The Middle AB
// martin@wehlou.com, martin@mitm.se, MiTM AB, Flintstensvagen 10, 75267 Uppsala, Sweden
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// $LastChangedDate: $
// $LastChangedRevision: $
// $LastChangedBy: $
// $HeadURL: $
// $Id: $
// ==============================================================


#include "stdafx.h"
#pragma warning(disable: 4786)
#include <sstream>
#include "exceptions.h"
#include "funcs.h"
#include "valstack.h"
#include "vstr.h"
#include "vset.h"
#include "vxml.h"
#include "constants.h"
#include "symbols.h"


using std::wstringstream;
using std::make_pair;



// ==========================================================================
//  CNodeVal
// ==========================================================================

CNodeVal::CNodeVal()
{}

CNodeVal::CNodeVal(XML_NodePtr spNode) :
m_spNode(spNode)
{}

CNodeVal::~CNodeVal()
{}

// ==========================================================================
//  CSymbol
// ==========================================================================

CSymbol::CSymbol(const wstring& sSymbolName, CVStr* pVStr) :
m_sSymbolName(sSymbolName),
m_vt(vtSTR),
m_pVStr(pVStr)
{}

CSymbol::CSymbol(const wstring& sSymbolName, double* pDbl) :
m_sSymbolName(sSymbolName),
m_vt(vtDBL),
m_pDbl(pDbl)
{}

CSymbol::CSymbol(const wstring& sSymbolName, bool* pBool) :
m_sSymbolName(sSymbolName),
m_vt(vtBOOL),
m_pBool(pBool)
{}

CSymbol::CSymbol(const wstring& sSymbolName, CVMap* pVMap) :
m_sSymbolName(sSymbolName),
m_vt(vtMAP),
m_pVMap(pVMap)
{}

CSymbol::CSymbol(const wstring& sSymbolName, CVSet* pVSet) :
m_sSymbolName(sSymbolName),
m_vt(vtSET),
m_pVSet(pVSet)
{}

CSymbol::CSymbol(const wstring& sSymbolName, CFuncVal* pFuncVal) :
m_sSymbolName(sSymbolName),
m_vt(vtFUNC),
m_pFuncVal(pFuncVal)
{}

CSymbol::CSymbol(const wstring& sSymbolName, CNodeVal* pNodeVal) :
m_sSymbolName(sSymbolName),
m_vt(vtNODE),
m_pNodeVal(pNodeVal)
{}

CSymbol::CSymbol(const wstring& sSymbolName, CNodelist* pNodelist) :
m_sSymbolName(sSymbolName),
m_vt(vtNODELIST),
m_pNodelist(pNodelist)
{}

CSymbol::CSymbol(const wstring& sSymbolName, CVXml* pVXml) :
m_sSymbolName(sSymbolName),
m_vt(vtXMLREF),
m_pVXml(pVXml)
{}

CSymbol::~CSymbol()
{
	switch(m_vt)
	{
	case vtSTR:			delete m_pVStr;			break;
	case vtDBL:			delete m_pDbl;			break;
	case vtBOOL:		delete m_pBool;			break;
	case vtMAP:			delete m_pVMap;			break;
	case vtSET:			delete m_pVSet;			break;
	case vtFUNC:		delete m_pFuncVal;		break;
	case vtNODE:		delete m_pNodeVal;		break;
	case vtNODELIST:	delete m_pNodelist;		break;
	case vtXMLREF:		delete m_pVXml;			break;

	default:
		// can't throw here, too bad, just don't forget any cases, duh!
		break;
	}
}

// ------------------------------------------------------------------------- //

void CSymbol::setStr(const wstring& sValue)
{
	if (m_vt != vtSTR)
		throw x::symbol_error(L"cannot set to string", m_sSymbolName);
	m_pVStr->get() = sValue;
}

// ------------------------------------------------------------------------- //

wstring CSymbol::getAsStr()
{
	if (m_vt == vtSTR)
		return m_pVStr->get();
	else
		throw x::symbol_error(L"could not convert to string", m_sSymbolName);
}

// ------------------------------------------------------------------------- //

double CSymbol::getAsDbl()
{
	if (m_vt == vtDBL)
		return *m_pDbl;
	else
		throw x::symbol_error(L"could not convert to double", m_sSymbolName);
}

// ------------------------------------------------------------------------- //

double& CSymbol::getDbl()
{
	if (m_vt != vtDBL)
		throw x::symbol_error(L"not a double", m_sSymbolName);
	return *m_pDbl;
}

// ------------------------------------------------------------------------- //

bool& CSymbol::getBool()
{
	if (m_vt != vtBOOL)
		throw x::symbol_error(L"not a bool", m_sSymbolName);
	return *m_pBool;
}

// ------------------------------------------------------------------------- //

CVStr& CSymbol::getVStr()
{
	if (m_vt != vtSTR)
		throw x::symbol_error(L"not a string", m_sSymbolName);
	return *m_pVStr;
}

// ------------------------------------------------------------------------- //

CVMap& CSymbol::getVMap()
{
	if (m_vt != vtMAP)
		throw x::symbol_error(L"not a map", m_sSymbolName);
	return *m_pVMap;
}

// ------------------------------------------------------------------------- //

CVSet& CSymbol::getVSet()
{
	if (m_vt != vtSET)
		throw x::symbol_error(L"not a set", m_sSymbolName);
	return *m_pVSet;
}

// ------------------------------------------------------------------------- //

CNodeVal& CSymbol::getNodeVal()
{
	if (m_vt != vtNODE)
		throw x::symbol_error(L"expected an XML node", m_sSymbolName);
	return *m_pNodeVal;
}

// ------------------------------------------------------------------------- //

CNodelist& CSymbol::getNodelist()
{
	if (m_vt != vtNODELIST)
		throw x::symbol_error(L"expected a nodelist", m_sSymbolName);
	return *m_pNodelist;
}

// ------------------------------------------------------------------------- //

CVXml& CSymbol::getVXml()
{
	if (m_vt != vtXMLREF)
		throw x::symbol_error(L"expected an XML reference", m_sSymbolName);
	return *m_pVXml;
}

// ------------------------------------------------------------------------- //

CFuncVal& CSymbol::getFuncVal()
{
	if (m_vt != vtFUNC)
		throw x::symbol_error(L"expected a function", m_sSymbolName);
	return *m_pFuncVal;
}

// ==========================================================================
//  CSymbols
// ==========================================================================

CSymbols::CSymbols() : m_lScope(0), m_pRootSymbols(NULL)
{
	// set up first scope
	m_vmSyms.push_back(t_mStrSym());
}

CSymbols::~CSymbols()
{}

// ------------------------------------------------------------------------- //

void CSymbols::defineStr(const wstring& sVarName, const wstring& sValue)
{
	verifyNotDefined(sVarName);
	CSymbolRcp rcpSymbol(new CSymbol(sVarName, new CVStr(sValue)));
	m_vmSyms[m_lScope].insert(make_pair(sVarName, rcpSymbol));
}

// ------------------------------------------------------------------------- //

void CSymbols::defineDbl(const wstring& sVarName, double dbl)
{
	verifyNotDefined(sVarName);
	CSymbolRcp rcpSymbol(new CSymbol(sVarName, new double(dbl)));
	m_vmSyms[m_lScope].insert(make_pair(sVarName, rcpSymbol));
}

// ------------------------------------------------------------------------- //

void CSymbols::defineBool(const wstring& sVarName, bool bValue)
{
	verifyNotDefined(sVarName);
	CSymbolRcp rcpSymbol(new CSymbol(sVarName, new bool(bValue)));
	m_vmSyms[m_lScope].insert(make_pair(sVarName, rcpSymbol));
}

// ------------------------------------------------------------------------- //

void CSymbols::defineXMLRef(const wstring& sVarName, const wstring& sFileName)
{
	verifyNotDefined(sVarName);
	CSymbolRcp rcpSymbol(new CSymbol(sVarName, new CVXml(sFileName)));
	m_vmSyms[m_lScope].insert(make_pair(sVarName, rcpSymbol));
}

// ------------------------------------------------------------------------- //

CNodeVal& CSymbols::defineNode(const wstring& sVarName)
{
	verifyNotDefined(sVarName);
	CNodeVal* pNV = new CNodeVal();
	CSymbolRcp rcpSymbol(new CSymbol(sVarName, pNV));
	m_vmSyms[m_lScope].insert(make_pair(sVarName, rcpSymbol));
	return *pNV;
}

// ------------------------------------------------------------------------- //

void CSymbols::defineNode(const wstring& sVarName, XML_NodePtr spNode)
{
	verifyNotDefined(sVarName);
	CNodeVal* pNV = new CNodeVal(spNode);
	CSymbolRcp rcpSymbol(new CSymbol(sVarName, pNV));
	m_vmSyms[m_lScope].insert(make_pair(sVarName, rcpSymbol));
}

// ------------------------------------------------------------------------- //

CNodelist& CSymbols::defineNodelist(const wstring& sVarName)
{
	verifyNotDefined(sVarName);
	CNodelist* pNL = new CNodelist();
	CSymbolRcp rcpSymbol(new CSymbol(sVarName, pNL));
	m_vmSyms[m_lScope].insert(make_pair(sVarName, rcpSymbol));
	return *pNL;
}

// ------------------------------------------------------------------------- //

void CSymbols::defineNodelist(const wstring& sVarName, const CNodelist& nodelist)
{
	verifyNotDefined(sVarName);
	CNodelist* pNL = new CNodelist(nodelist);
	CSymbolRcp rcpSymbol(new CSymbol(sVarName, pNL));
	m_vmSyms[m_lScope].insert(make_pair(sVarName, rcpSymbol));
}

// ------------------------------------------------------------------------- //

CVMap& CSymbols::defineMap(const wstring& sVarName)
{
	verifyNotDefined(sVarName);
	CVMap* pM = new CVMap();
	CSymbolRcp rcpSymbol(new CSymbol(sVarName, pM));
	m_vmSyms[m_lScope].insert(make_pair(sVarName, rcpSymbol));
	return *pM;
}

// ------------------------------------------------------------------------- //

CVSet& CSymbols::defineSet(const wstring& sVarName)
{
	verifyNotDefined(sVarName);
	CVSet* pS = new CVSet();
	CSymbolRcp rcpSymbol(new CSymbol(sVarName, pS));
	m_vmSyms[m_lScope].insert(make_pair(sVarName, rcpSymbol));
	return *pS;
}

// ------------------------------------------------------------------------- //

CFuncVal& CSymbols::defineFunc(const wstring& sVarName, CFunctorBaseRcp rcpFunctor)
{
	verifyNotDefined(sVarName);
	CFuncVal* pF = new CFuncVal(rcpFunctor);
	CSymbolRcp rcpSymbol(new CSymbol(sVarName, pF));
	m_vmSyms[m_lScope].insert(make_pair(sVarName, rcpSymbol));
	return *pF;
}

// ------------------------------------------------------------------------- //

// used when closing files, for instance, undefs first found from inner scope
// to outer
void CSymbols::undefine(const wstring& sVarName)
{
	for (long lScope = m_lScope; lScope >= 0; --lScope)
	{
		t_mStrSym::iterator itr = m_vmSyms[lScope].find(sVarName);
		if (itr != m_vmSyms[lScope].end())
		{
			m_vmSyms[lScope].erase(itr);
			return;
		}
	}
	throw x::symbol_error(L"variable not found", sVarName);
}

// ------------------------------------------------------------------------- //

double CSymbols::getAsDbl(const wstring& sVarName)
{
	return getSymbol(sVarName)->getAsDbl();
}

// ------------------------------------------------------------------------- //

bool& CSymbols::getBool(const wstring& sVarName)
{
	return getSymbol(sVarName)->getBool();
}

// ------------------------------------------------------------------------- //

double& CSymbols::getDbl(const wstring& sVarName)
{
	return getSymbol(sVarName)->getDbl();
}

// ------------------------------------------------------------------------- //

CVStr& CSymbols::getVStr(const wstring& sVarName)
{
	return getSymbol(sVarName)->getVStr();
}

// ------------------------------------------------------------------------- //

CNodeVal& CSymbols::getNodeVal(const wstring& sVarName)
{
	return getSymbol(sVarName)->getNodeVal();
}

// ------------------------------------------------------------------------- //

CNodelist& CSymbols::getNodelist(const wstring& sVarName)
{
	return getSymbol(sVarName)->getNodelist();
}

// ------------------------------------------------------------------------- //

CVMap& CSymbols::getVMap(const wstring& sVarName)
{
	return getSymbol(sVarName)->getVMap();
}

// ------------------------------------------------------------------------- //

CVSet& CSymbols::getVSet(const wstring& sVarName)
{
	return getSymbol(sVarName)->getVSet();
}

// ------------------------------------------------------------------------- //

CSymbolRcp CSymbols::getSymbol(const wstring& sVarName)
{
	// search scopes from inner to outer
	for (long lScope = m_lScope; lScope >= 0; --lScope)
	{
		t_mStrSym::iterator itr = m_vmSyms[lScope].find(sVarName);
		if (itr != m_vmSyms[lScope].end())
			return itr->second;
	}

	// if this is run as a function parser/symbol table, it has a
	// pointer to the symbol table of the root parser. From that one,
	// we should look for global symbols like builtin functions. All
	// that stuff is in scope 0 in the root symbol table.
	if (m_pRootSymbols)				
	{
		t_mStrSym::iterator itr = m_pRootSymbols->m_vmSyms[0].find(sVarName);
		if (itr != m_pRootSymbols->m_vmSyms[0].end())
			return itr->second;
	}
	throw x::symbol_error(L"undefined symbol", sVarName);

}

// ------------------------------------------------------------------------- //

CVXml& CSymbols::getVXml(const wstring& sVarName)
{
	return getSymbol(sVarName)->getVXml();
}

// ------------------------------------------------------------------------- //

CFuncVal& CSymbols::getFuncVal(const wstring& sVarName)
{
	return getSymbol(sVarName)->getFuncVal();
}

// ------------------------------------------------------------------------- //

eValType CSymbols::getValType(const wstring& sVarName)
{
	return getSymbol(sVarName)->getValType();
}

// ------------------------------------------------------------------------- //

bool CSymbols::isDefined(const wstring& sVarName)
{
	// check any scope
	for (long lScope = m_lScope; lScope >= 0; --lScope)
	{
		t_mStrSym::iterator itr = m_vmSyms[lScope].find(sVarName);
		if (itr != m_vmSyms[lScope].end())
			return true;
	}
	return false;
}

// ------------------------------------------------------------------------- //

void CSymbols::verifyNotDefined(const wstring& sVarName)
{
	// check only current scope
	if (m_vmSyms[m_lScope].find(sVarName) != m_vmSyms[m_lScope].end())
		throw x::symbol_error(L"doubly defined symbol", sVarName);
}

// ------------------------------------------------------------------------- //

void CSymbols::addScope()
{
	m_vmSyms.push_back(t_mStrSym());
	m_lScope++;
}

// ------------------------------------------------------------------------- //

void CSymbols::dropScope()
{
	m_vmSyms.pop_back();
	m_lScope--;
}

// ------------------------------------------------------------------------- //

void CSymbols::setRootSymbols(CSymbols* pRootSymbols)
{
	m_pRootSymbols = pRootSymbols;
}

// ------------------------------------------------------------------------- //
