// ==============================================================
// Name:		symbols.h
//
// Description:	x2c - symbol table
//				note: there are several symbol tables, one for each nesting level
//
// Copyright (C) 2001-2008, J. Martin Wehlou, Man in The Middle AB
// martin@wehlou.com, martin@mitm.se, MiTM AB, Flintstensvagen 10, 75267 Uppsala, Sweden
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// $LastChangedDate: $
// $LastChangedRevision: $
// $LastChangedBy: $
// $HeadURL: $
// $Id: $
// ==============================================================


#ifndef SYMBOLS_H
#define SYMBOLS_H
#pragma message("--- symbols.h")

#pragma message("- begin includes - " __FILE__)
#include <string>
#include <map>
#include "rcptr.h"
#include "MSXMLImport.h"
#include "nodelist.h"
#include "valstack.h"
#include "lex.h"
#include "vfunc.h"
#include "constants.h"
#include <iostream>
#include <fstream>
#pragma message("- end includes - " __FILE__)

//using std::string;
using std::wstring;
using std::map;
using std::streampos;
using std::ofstream;


//
//	so, what are we doing here? Well, we're having this container called CSymbols
//	that's mainly a map holding reference counting pointers to CSymbol instances. 
//	Having this reference counted saves us buckets of grief, since the map keeps using
//	assignment and copy constructors to its hearts delight.
//	Each symbol, however, can be a lot of mutually exclusive things, like functions, variables
//	and filereferences. Each with their own characteristics, dreams and desires.
//	Deriving all from a common base would only mean you'd need to have *all* their disparate
//	functionality in pure abstract functions in the base, then stub most of them out in the
//	children. And we all know how much work it can be to stub out anything in children. So
//	we won't do that here. Instead, we'll have a class containing the pointer to the relevant
//	implementation class. Having these pointers as a union means it doesn't waste space.
//

typedef map<wstring,wstring>	t_mStrStr;
typedef map<double,wstring>		t_mDblStr;

// ------------------------------------------------------------------------- //

class CVSet;
class CVXml;
class CVStr;
class CSymbols;
class CParse;


// ------------------------------------------------------------------------- //

class CNodeVal
{
public:
	CNodeVal();
	CNodeVal(XML_NodePtr spNode);
	virtual ~CNodeVal();
	XML_NodePtr getNode()		{ return m_spNode; }
private:
	XML_NodePtr		m_spNode;
};

// ------------------------------------------------------------------------- //

class CSymbol : public rcBase<CSymbol>
{
public:
	CSymbol(const wstring& sSymbolName, CVStr* pVStr);
	CSymbol(const wstring& sSymbolName, double* pDbl);
	CSymbol(const wstring& sSymbolName, bool* pBool);
	CSymbol(const wstring& sSymbolName, CVMap* pVMap);
	CSymbol(const wstring& sSymbolName, CVSet* pVSet);
	CSymbol(const wstring& sSymbolName, CFuncVal* pFuncVal);
	CSymbol(const wstring& sSymbolName, CNodeVal* pNodeVal);
	CSymbol(const wstring& sSymbolName, CNodelist* pNodelist);
	CSymbol(const wstring& sSymbolName, CVXml* pVXml);
	virtual			~CSymbol();
	void			setStr(const wstring& sValue);
	wstring			getAsStr();
	double			getAsDbl();
	double&			getDbl();
	bool&			getBool();
	CVStr&			getVStr();
	CVMap&			getVMap();
	CVSet&			getVSet();
	CNodeVal&		getNodeVal();
	CNodelist&		getNodelist();
	CVXml&			getVXml();
	CFuncVal&		getFuncVal();
	eValType		getValType() const		{ return m_vt; }
private:
	wstring			m_sSymbolName;
	eValType		m_vt;
	union	
	{
		CVStr*			m_pVStr;
		double*			m_pDbl;
		bool*			m_pBool;
		CVMap*			m_pVMap;
		CVSet*			m_pVSet;
		CFuncVal*		m_pFuncVal;
		CNodeVal*		m_pNodeVal;
		CNodelist*		m_pNodelist;
		CVXml*			m_pVXml;
	};
};
typedef rcPtr<CSymbol> CSymbolRcp;

// ------------------------------------------------------------------------- //

typedef map<wstring, CSymbolRcp>	t_mStrSym;
typedef	vector<t_mStrSym>			t_vmStrSym;

class CSymbols
{
public:
	CSymbols();
	~CSymbols();
	void			defineDbl		(const wstring& sVarName, double dblValue = 0.0);
	void			defineStr		(const wstring& sVarName, const wstring& sValue = L"");
	void			defineBool		(const wstring& sVarName, bool bValue);
	void			defineXMLRef	(const wstring& sVarName, const wstring& sFileName);
	CNodeVal&		defineNode		(const wstring& sVarName);
	void			defineNode		(const wstring& sVarName, XML_NodePtr spNode);
	CNodelist&		defineNodelist	(const wstring& sVarName);
	void			defineNodelist	(const wstring& sVarName, const CNodelist& nodelist);
	CFuncVal&		defineFunc		(const wstring& sVarName, CFunctorBaseRcp rcpFunctor);
//	CFuncVal&		defineFunc		(const wstring& sVarName, eValType vtReturns, int iParams, CFunctorBaseRcp rcpFunctor);
	CVMap&			defineMap		(const wstring& sVarName);
	CVSet&			defineSet		(const wstring& sVarName);
	void			undefine		(const wstring& sVarName);
	double			getAsDbl		(const wstring& sVarName);
	bool&			getBool			(const wstring& sVarName);
	double&			getDbl			(const wstring& sVarName);
	CVStr&			getVStr			(const wstring& sVarName);
	CVMap&			getVMap			(const wstring& sVarName);
	CVSet&			getVSet			(const wstring& sVarName);
	CNodeVal&		getNodeVal		(const wstring& sVarName);
	CNodelist&		getNodelist		(const wstring& sVarName);
	CVXml&			getVXml			(const wstring& sVarName);
	CFuncVal&		getFuncVal		(const wstring& sVarName);
	eValType		getValType		(const wstring& sVarName);

	void			addScope		();
	void			dropScope		();

	void			setRootSymbols	(CSymbols* pRootSymbols);

protected:
	CSymbolRcp		getSymbol		(const wstring& sVarName);
	bool			isDefined		(const wstring& sVarName);
	void			verifyNotDefined(const wstring& sVarName);
private:
	t_vmStrSym		m_vmSyms;
	long			m_lScope;
	CSymbols*		m_pRootSymbols;
};

// ----------------------------------------------------------------------- //
#else
#pragma message("--* symbols.h already included")
#endif
// ----------------------------------------------------------------------- //
