// ==============================================================
// Name:		x2c - valstack.cpp
//
// Description:	value stack
//
// Copyright (C) 2001-2008, J. Martin Wehlou, Man in The Middle AB
// martin@wehlou.com, martin@mitm.se, MiTM AB, Flintstensvagen 10, 75267 Uppsala, Sweden
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// $LastChangedDate: $
// $LastChangedRevision: $
// $LastChangedBy: $
// $HeadURL: $
// $Id: $
// ==============================================================

#include "stdafx.h"
#include "exceptions.h"
#include <iostream>			// debug dump
#include <sstream>
#include "valstack.h"

using std::cout;			// debug dump
using std::wcout;
using std::endl;			// debug dump
using std::stringstream;
using std::wstring;
using std::wstringstream;

// ==========================================================================
//  class CValStack::val
// ==========================================================================

CValStack::val::val() :
m_vt(vtNONE)
{}

CValStack::val::val(const val& src)
{
	copy(src);
}

CValStack::val& CValStack::val::operator =(const val& rhs)
{
	if (&rhs != this)
		copy(rhs);
	return *this;
}

void CValStack::val::copy(const val& src)
{
	m_vt = src.m_vt;
	switch(m_vt)
	{
	case vtSTR:			m_s = src.m_s;	break;
	case vtDBL:			m_dbl = src.m_dbl;	break;
	case vtBOOL:		m_b = src.m_b;	break;
	case vtNODE:		m_spNode = src.m_spNode; break;
	case vtNODELIST:	m_rcNodelist = src.m_rcNodelist; break;
	case vtMAP:			m_rcMap = src.m_rcMap; break;
	case vtSET:			m_rcSet = src.m_rcSet; break;
	}
}

CValStack::val::~val()
{}

CValStack::val::val(const wstring& sValue) :
m_vt(vtSTR), m_s(sValue) {}

CValStack::val::val(double dblValue) :
m_vt(vtDBL), m_dbl(dblValue) {}

CValStack::val::val(bool bValue) :
m_vt(vtBOOL), m_b(bValue) {}

CValStack::val::val(XML_NodePtr spNode) :
m_vt(vtNODE), m_spNode(spNode) {}

CValStack::val::val(const CNodelist& nodelist) :
m_vt(vtNODELIST), m_rcNodelist(CNodelistRcp(new CNodelist(nodelist))) {}

CValStack::val::val(const CVMap& vMap) :
m_vt(vtMAP), m_rcMap(CVMapRcp(new CVMap(vMap))) {}

CValStack::val::val(const CVSet& vSet) :
m_vt(vtSET), m_rcSet(CVSetRcp(new CVSet(vSet))) {}

// ------------------------------------------------------------------------- //

double CValStack::val::getDbl() const
{
	double dblRV = 0.0;

	switch(m_vt)
	{
	case vtDBL:		dblRV = m_dbl;					break;
	case vtSTR:		dblRV = _wtof(m_s.c_str());		break;
	case vtBOOL:	dblRV = m_b ? 1.0 : 0.0;		break;
	default:		dblRV = 0.0;					break;
	}
	return dblRV;
}

wstring CValStack::val::getStr() const
{
	wstring sRV;
	switch(m_vt)
	{
	case vtSTR:		sRV = m_s;						break;
	case vtBOOL:	sRV = m_b ? L"true" : L"false";	break;
	case vtNODE:	sRV = m_spNode->Gettext();		break;
	case vtNODELIST:return L"";						break;
	case vtDBL:
		{
			wstringstream ss;
			ss << m_dbl;
			sRV = ss.str();
		}
		break;

	case vtMAP:
		{
			for (CVMap::iterator itr = m_rcMap->begin(); itr != m_rcMap->end(); ++itr)
			{
				if (sRV.empty())
					sRV += L"{";
				else
					sRV += L",";
				sRV += L"\"";
				sRV += itr.getKey();
				sRV += L"\":\"";
				sRV += itr.getValue();
				sRV += L"\"";
			}
			sRV += L"}";
		}
		break;

	case vtSET:
		{
			for (CVSet::iterator itr = m_rcSet->begin(); itr != m_rcSet->end(); ++itr)
			{
				if (sRV.empty())
					sRV = L"{";
				else
					sRV += L",";
				sRV += L"\"";
				sRV += *itr;
				sRV += L"\"";
			}
			sRV += L"}";
		}
		break;
	}
	return sRV;
}

bool CValStack::val::getBool() const
{
	bool bRV = false;
	switch(m_vt)
	{
	case vtBOOL:	bRV = m_b;						break;
	case vtSTR:		bRV = !m_s.empty();				break;
	case vtNODE:	bRV = (m_spNode.p != NULL);		break;
	case vtNODELIST:bRV = (!m_rcNodelist->empty());	break;
	}
	return bRV;
}

XML_NodePtr CValStack::val::getNode() const
{
	if (m_vt == vtNODE)
		return m_spNode;
	else
		return XML_NodePtr();
}

CNodelist CValStack::val::getNodelist() const
{
	if (m_vt == vtNODELIST)
		return *m_rcNodelist;
	else
		return CNodelist();
}

CVMap CValStack::val::getVMap() const
{
	if (m_vt == vtMAP)
		return *m_rcMap;
	else
		return CVMap();
}

CVSet CValStack::val::getVSet() const
{
	if (m_vt == vtSET)
		return *m_rcSet;
	else
		return CVSet();
}

// ==========================================================================
//	class CValStack  
// ==========================================================================

CValStack::CValStack() {}

CValStack::~CValStack() {}

// ------------------------------------------------------------------------- //

void CValStack::push(const wstring& sValue)
{
	m_stack.push_back(val(sValue));
	dump();
}

void CValStack::push(double dblValue)
{
	m_stack.push_back(val(dblValue));
	dump();
}

void CValStack::push(bool bValue)
{
	m_stack.push_back(val(bValue));
	dump();
}

void CValStack::push(XML_NodePtr spNode)
{
	m_stack.push_back(val(spNode));
	dump();
}

void CValStack::push(const CNodelist& nodelist)
{
	m_stack.push_back(val(nodelist));
	dump();
}

void CValStack::push(const CVMap& vMap)
{
	m_stack.push_back(val(vMap));
	dump();
}

void CValStack::push(const CVSet& vSet)
{
	m_stack.push_back(val(vSet));
	dump();
}

// ------------------------------------------------------------------------- //

bool CValStack::isStr() const
{
	return getValType() == vtSTR;
}

bool CValStack::isDbl() const
{
	return getValType() == vtDBL;
}

bool CValStack::isBool() const
{
	return getValType() == vtBOOL;
}

bool CValStack::isNode() const
{
	return getValType() == vtNODE;
}

bool CValStack::isNodelist() const
{
	return getValType() == vtNODELIST;
}

bool CValStack::isVMap() const
{
	return getValType() == vtMAP;
}

bool CValStack::isVSet() const
{
	return getValType() == vtSET;
}

eValType CValStack::getValType() const
{
	if (m_stack.size() < 1) throw x::stackunderflow();
	return m_stack.back().getValType();
}

int CValStack::getStackSize() const
{
	return m_stack.size();
}

// ------------------------------------------------------------------------- //

void CValStack::pop()
{
	if (m_stack.size() < 1) throw x::stackunderflow();
	m_stack.pop_back();
}

wstring CValStack::popStr()
{
	if (m_stack.size() < 1) throw x::stackunderflow();
	wstring sRV = m_stack.back().getStr();;
	m_stack.pop_back();
	return sRV;
}

double CValStack::popDbl()
{
	if (m_stack.size() < 1) throw x::stackunderflow();
	double dblRV = m_stack.back().getDbl();
	m_stack.pop_back();
	return dblRV;
}

// whatever type the value has, pop it as a boolean
bool CValStack::popBool()
{
	if (m_stack.size() < 1) throw x::stackunderflow();
	bool bRV = m_stack.back().getBool();
	m_stack.pop_back();
	return bRV;
}

XML_NodePtr CValStack::popNode()
{
	if (m_stack.size() < 1) throw x::stackunderflow();
	XML_NodePtr spNode = m_stack.back().getNode();
	m_stack.pop_back();
	return spNode;
}

CNodelist CValStack::popNodelist()
{
	if (m_stack.size() < 1) throw x::stackunderflow();
	CNodelist nodelist = m_stack.back().getNodelist();
	m_stack.pop_back();
	return nodelist;
}

CVMap CValStack::popVMap()
{
	if (m_stack.size() < 1) throw x::stackunderflow();
	CVMap vMap = m_stack.back().getVMap();
	m_stack.pop_back();
	return vMap;
}

CVSet CValStack::popVSet()
{
	if (m_stack.size() < 1) throw x::stackunderflow();
	CVSet vSet = m_stack.back().getVSet();
	m_stack.pop_back();
	return vSet;
}

// ------------------------------------------------------------------------- //

void CValStack::dump()
{
	if (k_bValStackDump)
	{
		wcout << endl;
		wcout << L"value stack dump" << endl;
		wcout << L"================" << endl;
		for (vector<val>::size_type i = 0; i < m_stack.size(); i++)
		{
			switch (m_stack[i].getValType())
			{
			case vtSTR:		wcout << m_stack[i].getStr() << endl;	break;
			case vtDBL:		wcout << m_stack[i].getDbl() << endl;	break;
			case vtBOOL:	wcout << m_stack[i].getBool() << endl;	break;
			case vtNODE:	wcout << L"XML Node" << endl;			break;
			case vtNODELIST:wcout << L"Nodelist" << endl;			break;
			case vtMAP:		wcout << L"Map" << endl;				break;
			case vtSET:		wcout << L"Set" << endl;				break;
			}
		}
	}
}